{ *************************************************************************** }
{ TMS Cryptography Pack }
{ for Delphi & C++Builder }
{ }
{ Written by Cyberens Technologies & Services }
{ Exclusively commercialized by tmssoftware.com bvba }
{ copyright  2016 - 2020 by Cyberens Technologies & Services }
{ Email : info@tmssoftware.com }
{ Web : http://www.tmssoftware.com }
{ }
{ The source code is given as is. The author is not responsible }
{ for any possible damage done due to the use of this code. }
{ The component can be freely used in any application. The complete }
{ source code remains property of the author and may not be distributed, }
{ published, given or sold in any form as such. No parts of the source }
{ code can be included in any other component or application without }
{ written authorization of the author. }
{ *************************************************************************** }

unit HashObj;

interface

{$I TMSDEFS.INC}
{$I TMSCRYPTO.INC}

uses
  CryptBase, MiscObj, Classes;

{$POINTERMATH ON}

type
  TSHA256Key = array [0 .. 63] of Cardinal;
  TSHA512Key = array [0 .. 79] of Int64;
  TSHA256H = array [0 .. 7] of Cardinal;
  TSHA512H = array [0 .. 7] of Int64;
  TSHA3Array = array [0 .. 24] of Int64;
  TRIPEMDBuf = array [0 .. 4] of Integer;
  TRIPEMDX = array [0 .. 15] of Integer;
  PPByte = ^PByte;
  THashFunction = (hsha2, hsha3);
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  saltPB = array [0 .. 60] of AnsiChar;
  saltArg = PAnsiChar;
  TSHA256Block = array [0 .. 63] of AnsiChar;
  TSHA512Block = array [0 .. 127] of AnsiChar;

  Pblake2b_state = ^_blake2b_state;

  _blake2b_state = packed record
    hash_state: array [0 .. 8] of LongInt;
    number_bytes: array [0 .. 2] of LongInt;
    f: array [0 .. 2] of LongInt;
    buf: array [0 .. 128] of PAnsiChar;
    bufSizeBytes: Integer;
    outputSizeBytes: Integer;
    last_node: PAnsiChar;
  end;
{$ELSE}

  saltPB = array [0 .. 60] of Byte;
  saltArg = PByte;
  TSHA256Block = array [0 .. 63] of Byte;
  TSHA512Block = array [0 .. 127] of Byte;

  Pblake2b_state = ^_blake2b_state;

  _blake2b_state = packed record
    hash_state: array [0 .. 8] of LongInt;
    number_bytes: array [0 .. 2] of LongInt;
    f: array [0 .. 2] of LongInt;
    buf: array [0 .. 128] of PByte;
    bufSizeBytes: Integer;
    outputSizeBytes: Integer;
    last_node: PByte;
  end;
{$IFEND}

  TSHA2Hash = class(TTMSCryptBase)
  private
    FHashSizeBits: Integer;
    FOutputFormat: TConvertType;
    FUni: TUnicode;
    FProgress: Integer;
    FOnChange: TNotifyEvent;
    procedure SetHashSizeBits(const Value: Integer);
    procedure SetProgress(P: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(hashSizeBits: Integer; outputFormat: TConvertType;
      uni: TUnicode); reintroduce; overload;
    function Hash(s: string): string;
    function HashFile(s: string): string;
    function HashStream(s: TStream): string;
    function HMAC(s, k: string): string;
    function VerifyHMAC(s, k, h: string): Integer;
  published
    property hashSizeBits: Integer read FHashSizeBits write SetHashSizeBits
      default 256;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property Unicode: TUnicode read FUni write FUni default yesUni;
    property Progress: Integer read FProgress write SetProgress default 0;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TSHA2Hash }

  TSHA3Type = (tsha, txof);

  TSHA3Hash = class(TTMSCryptBase)
  private
    FHashSizeBits: Integer;
    FType: TSHA3Type;
    FVersion: Integer;
    FUni: TUnicode;
    FOutputFormat: TConvertType;
    FProgress: Integer;
    FOnChange: TNotifyEvent;
    procedure SetHashSizeBits(const Value: Integer);
    procedure SetType(const Value: TSHA3Type);
    procedure SetVersion(const Value: Integer);
    procedure SetProgress(P: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(hashSizeBits: Integer; outputFormat: TConvertType;
      uni: TUnicode); reintroduce; overload;
    Constructor Create(hashSizeBits: Integer; outputFormat: TConvertType;
      uni: TUnicode; version: Integer); reintroduce; overload;
    function Hash(s: string): string;
    function HashFile(s: string): string;
    function HashStream(s: TStream): string;
    function cSHAKEHash(s, salt: string): string;
    function KMACHash(k, s, salt: string): string;
    function TupleHash(s: array of string; salt: string): string;
    function HMAC(s, k: string): string;
    function VerifyHMAC(s, k, h: string): Integer;
  published
    property AType: TSHA3Type read FType write SetType default tsha;
    property hashSizeBits: Integer read FHashSizeBits write SetHashSizeBits
      default 256;
    property version: Integer read FVersion write SetVersion default 256;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property Unicode: TUnicode read FUni write FUni default yesUni;
    property Progress: Integer read FProgress write SetProgress default 0;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TSHA3Hash }

  TBlake2BHash = class(TTMSCryptBase)
  private
    FOutputFormat: TConvertType;
    FKey: String;
    FKeylen: Integer;
    FHashSizeBytes: Integer;
    FUni: TUnicode;
    FProgress: Integer;
    FOnChange: TNotifyEvent;
    procedure SetHashSizeBytes(const Value: Integer);
    procedure SetKey(const Value: String);
    procedure SetProgress(P: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(hashSizeBytes: Integer; key: string;
      outputFormat: TConvertType; uni: TUnicode); reintroduce; overload;
    function Hash(s: string): string;
    function HashFile(s: string): string;
    function HashStream(s: TStream): string;
  published
    property hashSizeBytes: Integer read FHashSizeBytes write SetHashSizeBytes
      default 16;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property key: String read FKey write SetKey;
    property Unicode: TUnicode read FUni write FUni default yesUni;
    property Progress: Integer read FProgress write SetProgress default 0;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TBlake2BHash }

  TRIPEMD160Hash = class(TTMSCryptBase)
  private
    FOutputFormat: TConvertType;
    FUni: TUnicode;
    FProgress: Integer;
    FOnChange: TNotifyEvent;
    procedure SetProgress(P: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(outputFormat: TConvertType; uni: TUnicode);
      reintroduce; overload;
    function Hash(s: string): string;
    function HashFile(s: string): string;
    function HashStream(s: TStream): string;
  published
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property Unicode: TUnicode read FUni write FUni default yesUni;
    property Progress: Integer read FProgress write SetProgress default 0;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TRIPEMD160Hash }

  TPBKDF2KeyDerivationOLD = class(TTMSCryptBase)
  private
    FOutputSizeBits: Integer;
    FSalt: saltPB;
    FStringSalt: string;
    FCounter: Integer;
    FUni: TUnicode;
    FOutputFormat: TConvertType;
    procedure SetOutputSizeBits(const Value: Integer);
    procedure SetStringSalt(const Value: string);
    procedure SetCounter(const Value: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(outputSizeBits: Integer; salt: string; counter: Integer;
      outputFormat: TConvertType; uni: TUnicode); reintroduce; overload;
    function GenerateKey(s: string): string;
  published
    property outputSizeBits: Integer read FOutputSizeBits
      write SetOutputSizeBits;
    property StringSalt: string read FStringSalt write SetStringSalt;
    property counter: Integer read FCounter write SetCounter;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat;
    property Unicode: TUnicode read FUni write FUni;
  end;

  { TPBKDF2KeyDerivation }

  TPBKDF2KeyDerivation = class(TTMSCryptBase)
  private
    FOutputSizeBits: Integer;
    FSalt: string;
    FCounter: Integer;
    FUni: TUnicode;
    FOutputFormat: TConvertType;
    FHashFunction: THashFunction;
    FHashSizeBits: Integer;
    procedure SetOutputSizeBits(const Value: Integer);
    procedure SetHashSizeBits(const Value: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(outputSizeBits: Integer; salt: string; counter: Integer;
      outputFormat: TConvertType; uni: TUnicode; hashF: THashFunction;
      hashSB: Integer); reintroduce; overload;
    function GenerateKey(s: string): string;
  published
    property outputSizeBits: Integer read FOutputSizeBits
      write SetOutputSizeBits default 128;
    property salt: string read FSalt write FSalt;
    property counter: Integer read FCounter write FCounter default 10000;
    property hashFunction: THashFunction read FHashFunction write FHashFunction
      default hsha2;
    property hashSizeBits: Integer read FHashSizeBits write SetHashSizeBits
      default 256;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property Unicode: TUnicode read FUni write FUni default yesUni;
  end;

  { TPBKDF2KeyDerivation }

  TArgon2KeyDerivation = class(TTMSCryptBase)
  private
    FOutputSizeBytes: Integer;
    FMemory: Integer;
    FSalt: saltArg;
    FStringSalt: string;
    FCounter: Integer;
    FUni: TUnicode;
    FOutputFormat: TConvertType;
    procedure SetOutputSizeBytes(const Value: Integer);
    procedure SetStringSalt(const Value: string);
    procedure SetCounter(const Value: Integer);
    procedure SetMemory(const Value: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(outputSizeBytes: Integer; salt: string; counter: Integer;
      outputFormat: TConvertType; memory: Integer; uni: TUnicode);
      reintroduce; overload;
    destructor Destroy; override;
    function GenerateKey(s: string): string;
  published
    property outputSizeBytes: Integer read FOutputSizeBytes
      write SetOutputSizeBytes default 16;
    property StringSalt: string read FStringSalt write SetStringSalt;
    property counter: Integer read FCounter write SetCounter default 10;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property memory: Integer read FMemory write SetMemory default 16;
    property Unicode: TUnicode read FUni write FUni default yesUni;
  end;

  { TArgon2KeyDerivation }

  THKDFKeyDerivation = class(TTMSCryptBase)
  private
    FUni: TUnicode;
    FOutputFormat: TConvertType;
    FHashFunction: THashFunction;
    FHashSizeBits: Integer;
    procedure SetHashSizeBits(const Value: Integer);
  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(outputFormat: TConvertType; uni: TUnicode;
      hashF: THashFunction; hashSB: Integer); reintroduce; overload;
    function Expand(s, info: string; len: Integer): string;
    function Extract(s, salt: string): string;
  published
    property hashFunction: THashFunction read FHashFunction write FHashFunction
      default hsha2;
    property hashSizeBits: Integer read FHashSizeBits write SetHashSizeBits
      default 256;
    property outputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property Unicode: TUnicode read FUni write FUni default yesUni;
  end;

  { THKDFKeyDerivation }

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

function SHA3_XOF_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PAnsiChar)
  : Integer; cdecl;
function SHA3_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
function SHA3_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl;
function SHA3_XOF_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  version: Integer; fileHash: PAnsiChar): Integer; cdecl;
function SHA3_cSHAKE_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PAnsiChar; saltSizeBytes: Int64;
  bufferHash: PAnsiChar; version: Integer): Integer; cdecl;
function SHA3_KMAC_BufferHash(key: PAnsiChar; keySizeBytes: Int64;
  inputBuffer: PAnsiChar; inputSizeBytes: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl;
function SHA3_TupleHASH_BufferHash(inputBuffer: PPAnsiChar;
  inputSizeBytes: PInt64; nbInputBuffer: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl;
procedure KeccakF1600_StatePermute(var state: TSHA3Array); cdecl;
function Blake2B_FileHash(input_path: PAnsiChar; key: PAnsiChar;
  keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer): Integer; cdecl;
function Blake2B_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer)
  : Integer; cdecl;
function update_blake2b(s: Pblake2b_state; input: PAnsiChar;
  inputSizeBytes: Integer): Integer; cdecl;
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer)
  : Integer; cdecl;
function finalHash_blake2b(s: Pblake2b_state; outB: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer): Integer; cdecl;
procedure RIPEMD160_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  output: PAnsiChar); cdecl;
function RIPEMD160_FileHash(inputPath: PAnsiChar; output: PAnsiChar)
  : Integer; cdecl;
{$IFDEF MACOS64}
procedure init_ripemd160(buf: PInteger); cdecl;
procedure compress_ripemd160(buf: PInteger; X: PInteger); cdecl;
procedure finish_ripemd160(buf: PInteger; str: PAnsiChar; lswlen: Integer;
  mswlen: Integer); cdecl;
{$ELSE}
procedure init_ripemd160(var buf: TRIPEMDBuf); cdecl;
procedure compress_ripemd160(var buf: TRIPEMDBuf; X: TRIPEMDX); cdecl;
procedure finish_ripemd160(var buf: TRIPEMDBuf; str: PAnsiChar; lswlen: Integer;
  mswlen: Integer); cdecl;

{$ENDIF}
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PAnsiChar; pwdSizeBytes: Integer; salt: saltArg; output: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
function SHA2_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl;
function SHA1_FileHash(inputFilePath: PAnsiChar; fileHash: PAnsiChar)
  : Integer; cdecl;
function SHA2_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
function SHA1_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  bufferHash: PAnsiChar): Integer; cdecl;
{$IFDEF MACOS64}
procedure SHA256_BlockHash(mess: PAnsiChar; k: PCardinal; h: PCardinal); cdecl;
procedure SHA512_BlockHash(mess: PAnsiChar; k: PInt64; h: PInt64); cdecl;
procedure SHA384_BlockHash(mess: PAnsiChar; k: PInt64; h: PInt64); cdecl;
{$ELSE}
procedure SHA256_BlockHash(mess: TSHA256Block; k: TSHA256Key;
  var h: TSHA256H); cdecl;
procedure SHA512_BlockHash(mess: TSHA512Block; k: TSHA512Key;
  var h: TSHA512H); cdecl;
procedure SHA384_BlockHash(mess: TSHA512Block; k: TSHA512Key;
  var h: TSHA512H); cdecl;
{$ENDIF}
procedure Constants_sha2(var k: TSHA256Key); cdecl;
procedure Constants_sha512(var k: TSHA512Key); cdecl;
function SHA2_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl;
function SHA2_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl;
function SHA3_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl;
function SHA3_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl;
function SHA3_XOF_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;
function SHA3_XOF_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;

function PBKDF2_KeyDerivationFunction(password: PAnsiChar;
  inputSizeBytes: Integer; salt: PAnsiChar; saltSizeBytes: Integer;
  counter: Integer; h: Integer; hashSizeBits: Integer; outputSizeBits: Integer;
  output: PAnsiChar): Integer; cdecl;
function PBKDF2_KeyDerivationFunction_OLD(password: PAnsiChar; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PAnsiChar): Integer; cdecl;
{$ELSE}
function SHA3_XOF_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PByte): Integer;
function SHA3_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer;
function SHA3_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer;
function SHA3_XOF_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  version: Integer; fileHash: PByte): Integer;
function SHA3_cSHAKE_BufferHash(inputBuffer: PByte; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64; bufferHash: PByte;
  version: Integer): Integer;
function SHA3_KMAC_BufferHash(key: PByte; keySizeBytes: Int64;
  inputBuffer: PByte; inputSizeBytes: Int64; hashSizeBits: Int64; salt: PByte;
  saltSizeBytes: Int64; bufferHash: PByte; version: Integer): Integer;
function SHA3_TupleHASH_BufferHash(inputBuffer: PPByte; inputSizeBytes: PInt64;
  nbInputBuffer: Int64; hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64;
  bufferHash: PByte; version: Integer): Integer;
procedure KeccakF1600_StatePermute(var state: TSHA3Array);
function Blake2B_FileHash(input_path: PByte; key: PByte; keylen: Integer;
  output: PByte; outputSizeBytes: Integer): Integer;
function Blake2B_BufferHash(input: PByte; inputSizeBytes: Integer; key: PByte;
  keylen: Integer; output: PByte; outputSizeBytes: Integer): Integer;
function update_blake2b(s: Pblake2b_state; input: PByte;
  inputSizeBytes: Integer): Integer;
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer): Integer;
function finalHash_blake2b(s: Pblake2b_state; outB: PByte;
  outputSizeBytes: Integer): Integer;
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PByte; keylen: Integer): Integer;
procedure RIPEMD160_BufferHash(input: PByte; inputSizeBytes: Integer;
  output: PByte);
function RIPEMD160_FileHash(inputPath: PByte; output: PByte): Integer;
procedure init_ripemd160(var buf: TRIPEMDBuf);
procedure compress_ripemd160(var buf: TRIPEMDBuf; X: TRIPEMDX);
procedure finish_ripemd160(var buf: TRIPEMDBuf; str: PByte; lswlen: Integer;
  mswlen: Integer);
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PByte; pwdSizeBytes: Integer; salt: saltArg; output: PByte;
  outputSizeBytes: Integer): Integer;
function SHA2_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer;
function SHA1_FileHash(inputFilePath: PByte; fileHash: PByte): Integer;
function SHA2_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer;
function SHA1_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  bufferHash: PByte): Integer;
procedure SHA256_BlockHash(mess: TSHA256Block; k: TSHA256Key; var h: TSHA256H);
procedure SHA384_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
procedure SHA512_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
procedure Constants_sha2(var k: TSHA256Key);
procedure Constants_sha512(var k: TSHA512Key);
function SHA2_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
function SHA2_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer;
  HMAC: PByte): Integer;
function PBKDF2_KeyDerivationFunction_OLD(password: PByte; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PByte): Integer;
function SHA3_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
function SHA3_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer;
  HMAC: PByte): Integer;
function SHA3_XOF_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; version: Integer;
  HMAC: PByte): Integer;
function SHA3_XOF_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; version: Integer;
  HMAC: PByte): Integer;

function PBKDF2_KeyDerivationFunction(password: PByte; inputSizeBytes: Integer;
  salt: PByte; saltSizeBytes: Integer; counter: Integer; h: Integer;
  hashSizeBits: Integer; outputSizeBits: Integer; output: PByte): Integer;
{$IFEND}

implementation

uses
  SysUtils, Math,
{$IFDEF MSWINDOWS}
  Winapi.Windows, AnsiStrings, crtl,
{$ENDIF MSWINDOWS}
{$IFDEF POSIX}
  Posix.Dlfcn, Posix.Fcntl, Posix.SysStat, Posix.SysTime, Posix.SysTypes,
  Posix.Base, Posix.Unistd, Posix.Stdlib, Posix.Stdio,
{$ENDIF POSIX}
{$IFDEF PC_MAPPED_EXCEPTIONS}
  System.Internal.Unwinder,
{$ENDIF PC_MAPPED_EXCEPTIONS}
  System.SysConst;

const
  ADVAPI32 = 'advapi32.dll';
  msvcrt = 'msvcrt.dll';
{$IFNDEF WINXP}
  BCRYPT = 'Bcrypt.dll';
{$ENDIF}
{$IF not defined(IDEVERSION1021)}
  RANDOMDLL = 'RandomDLL.dll'; // Copy the DLL from Win64 directory
  // into C:\Windows\SysWOW64 if you are running 32 bit Windows
  // or into C:\Windows\System32 if you are running 64 bit Windows
{$IFEND}
  libc = '/usr/lib/libc.dylib';
  PROV_RSA_FULL = 1;
  CRYPT_VERIFYCONTEXT = $F0000000;
{$IF (defined(WIN32) or defined(MACOS)) and (not defined(IOS))}
  PU = '_';
{$ELSE}
  PU = '';
{$IFEND}
{$IF defined(WIN32)}
{$LINK './libWIN32/sha2.obj'}
{$LINK './libWIN32/sha3.obj'}
{$LINK './libWIN32/PBKDF2.obj'}
{$LINK './libWIN32/PBKDF2_old.obj'}
{$IFDEF WINXP}
{$LINK './libWIN32/elt_rep_XP.obj'}
{$ELSE}
{$LINK './libWIN32/elt_rep.obj'}
{$ENDIF}
{$LINK './libWIN32/error.obj'}
{$LINK './libWIN32/blake2b.obj'}
{$LINK './libWIN32/argon2d.obj'}
{$LINK './libWIN32/ripemd160.obj'}
{$ELSEIF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
{$LINK './libWIN64/sha2.o'}
{$LINK './libWIN64/sha3.o'}
{$LINK './libWIN64/PBKDF2.o'}
{$LINK './libWIN64/PBKDF2_old.o'}
{$LINK './libWIN64/elt_rep.o'}
{$LINK './libWIN64/error.o'}
{$LINK './libWIN64/blake2b.o'}
{$LINK './libWIN64/argon2d.o'}
{$LINK './libWIN64/ripemd160.o'}
{$ELSE}
{$LINK './libWIN64/sha2.obj'}
{$LINK './libWIN64/sha3.obj'}
{$LINK './libWIN64/PBKDF2.obj'}
{$LINK './libWIN64/PBKDF2_old.obj'}
{$LINK './libWIN64/elt_rep.obj'}
{$LINK './libWIN64/error.obj'}
{$LINK './libWIN64/blake2b.obj'}
{$LINK './libWIN64/argon2d.obj'}
{$LINK './libWIN64/ripemd160.obj'}
{$IFEND}
{$ELSEIF defined(MACOS32) and not defined(IOS)}
{$LINK './libOSX32/sha2.o'}
{$LINK './libOSX32/sha3.o'}
{$LINK './libOSX32/PBKDF2.o'}
{$LINK './libOSX32/PBKDF2_old.o'}
{$LINK './libOSX32/elt_rep.o'}
{$LINK './libOSX32/error.o'}
{$LINK './libOSX32/blake2b.o'}
{$LINK './libOSX32/argon2d.o'}
{$LINK './libOSX32/ripemd160.o'}
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS32)) and (not defined(IOS))}
function SHA3_XOF_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PAnsiChar): Integer;
  cdecl; external name PU + 'SHA3_XOF_BufferHash';
function SHA3_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
  external name PU + 'SHA3_BufferHash';
function SHA3_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl; external name PU + 'SHA3_FileHash';
function SHA3_XOF_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  version: Integer; fileHash: PAnsiChar): Integer; cdecl;
  external name PU + 'SHA3_XOF_FileHash';
function SHA3_cSHAKE_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PAnsiChar; saltSizeBytes: Int64;
  bufferHash: PAnsiChar; version: Integer): Integer; cdecl;
  external name PU + 'SHA3_cSHAKE_BufferHash';
function SHA3_KMAC_BufferHash(key: PAnsiChar; keySizeBytes: Int64;
  inputBuffer: PAnsiChar; inputSizeBytes: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl; external name PU + 'SHA3_KMAC_BufferHash';
function SHA3_TupleHASH_BufferHash(inputBuffer: PPAnsiChar;
  inputSizeBytes: PInt64; nbInputBuffer: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl;
  external name PU + 'SHA3_TupleHASH_BufferHash';
procedure KeccakF1600_StatePermute(var state: TSHA3Array); cdecl;
  external name PU + 'KeccakF1600_StatePermute';
function Blake2B_FileHash(input_path: PAnsiChar; key: PAnsiChar;
  keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer): Integer; cdecl;
  external name PU + 'Blake2B_FileHash';
function Blake2B_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer)
  : Integer; cdecl; external name PU + 'Blake2B_BufferHash';
function update_blake2b(s: Pblake2b_state; input: PAnsiChar;
  inputSizeBytes: Integer): Integer; cdecl; external name PU + 'update_blake2b';
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer): Integer;
  cdecl; external name PU + 'init_blake2b';
function finalHash_blake2b(s: Pblake2b_state; outB: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
  external name PU + 'finalHash_blake2b';
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer): Integer; cdecl;
  external name PU + 'initKey_blake2b';
procedure RIPEMD160_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  output: PAnsiChar); cdecl; external name PU + 'RIPEMD160_BufferHash';
function RIPEMD160_FileHash(inputPath: PAnsiChar; output: PAnsiChar): Integer;
  cdecl; external name PU + 'RIPEMD160_FileHash';
procedure init_ripemd160(var buf: TRIPEMDBuf); cdecl;
  external name PU + 'init_ripemd160';
procedure compress_ripemd160(var buf: TRIPEMDBuf; X: TRIPEMDX); cdecl;
  external name PU + 'compress_ripemd160';
procedure finish_ripemd160(var buf: TRIPEMDBuf; str: PAnsiChar; lswlen: Integer;
  mswlen: Integer); cdecl; external name PU + 'finish_ripemd160';
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PAnsiChar; pwdSizeBytes: Integer; salt: saltArg; output: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
  external name PU + 'Argon2_KeyDerivation';
function SHA2_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl; external name PU + 'SHA2_FileHash';
function SHA1_FileHash(inputFilePath: PAnsiChar; fileHash: PAnsiChar): Integer;
  cdecl; external name PU + 'SHA1_FileHash';
function SHA2_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
  external name PU + 'SHA2_BufferHash';
function SHA1_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  bufferHash: PAnsiChar): Integer; cdecl; external name PU + 'SHA1_BufferHash';
procedure SHA256_BlockHash(mess: TSHA256Block; k: TSHA256Key; var h: TSHA256H);
  cdecl; external name PU + 'SHA256_BlockHash';
procedure SHA512_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  cdecl; external name PU + 'SHA512_BlockHash';
procedure SHA384_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  cdecl; external name PU + 'SHA384_BlockHash';
procedure Constants_sha2(var k: TSHA256Key); cdecl;
  external name PU + 'Constants_sha2';
procedure Constants_sha512(var k: TSHA512Key); cdecl;
  external name PU + 'Constants_sha512';
function SHA2_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external name PU + 'SHA2_HMAC';
function SHA2_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external name PU + 'SHA2_VerifyHMAC';
function PBKDF2_KeyDerivationFunction_OLD(password: PAnsiChar; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PAnsiChar): Integer; cdecl;
  external name PU + 'PBKDF2_KeyDerivationFunction_OLD';
function SHA3_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external name PU + 'SHA3_HMAC';
function SHA3_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external name PU + 'SHA3_VerifyHMAC';
function SHA3_XOF_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;
  external name PU + 'SHA3_XOF_HMAC';
function SHA3_XOF_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;
  external name PU + 'SHA3_XOF_VerifyHMAC';
function PBKDF2_KeyDerivationFunction(password: PAnsiChar;
  inputSizeBytes: Integer; salt: PAnsiChar; saltSizeBytes: Integer;
  counter: Integer; h: Integer; hashSizeBits: Integer; outputSizeBits: Integer;
  output: PAnsiChar): Integer; cdecl;
  external name PU + 'PBKDF2_KeyDerivationFunction';
function CharToHexa(MyChar: PAnsiChar; LenChar: Integer; HexaChar: PAnsiChar)
  : Integer; cdecl; external name PU + 'CharToHexa';
function HexaToChar(HexaChar: PAnsiChar; MyChar: PAnsiChar): Integer; cdecl;
  external name PU + 'HexaToChar';
{$IF defined(WIN32) or defined(MACOS32)}
procedure _unsignedIntToTab(w: Integer; tab: PAnsiChar); cdecl; external;
function _tabToUnsignedLongLong(tab: PAnsiChar): LongInt; cdecl; external;
procedure _permutation_blake2b(v: PLongInt; vo: Integer; v1: Integer;
  v2: Integer; v3: Integer; v4: Integer; v5: Integer; v6: Integer; v7: Integer;
  v8: Integer; v9: Integer; v10: Integer; v11: Integer; v12: Integer;
  v13: Integer; v14: Integer; v15: Integer); cdecl; external;
procedure _unsignedLongLongToTab(w: LongInt; tab: PAnsiChar); cdecl; external;
{$ELSE}
procedure unsignedIntToTab(w: Integer; tab: PAnsiChar); cdecl; external;
function tabToUnsignedLongLong(tab: PAnsiChar): LongInt; cdecl; external;
procedure permutation_blake2b(v: PLongInt; vo: Integer; v1: Integer;
  v2: Integer; v3: Integer; v4: Integer; v5: Integer; v6: Integer; v7: Integer;
  v8: Integer; v9: Integer; v10: Integer; v11: Integer; v12: Integer;
  v13: Integer; v14: Integer; v15: Integer); cdecl; external;
procedure unsignedLongLongToTab(w: LongInt; tab: PAnsiChar); cdecl; external;
{$IFEND}
{$IFDEF MACOS32}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomUBuffer';
function RandomInt(): Integer; external name PU + 'RandomInt';
function _open(PathName: MarshaledAString; Flags: Integer): Integer; cdecl;
  varargs; external libc name PU + 'open';
function _read(Handle: Integer; Buffer: Pointer; Count: size_t): ssize_t; cdecl;
  external libc name PU + 'read';
function _close(filedes: Integer): Integer; cdecl;
  external libc name PU + 'close';
function _memset(var b; c: Integer; len: size_t): Pointer; cdecl;
  external libc name PU + 'memset';
function _memcpy(var b; var a; num: size_t): Pointer; cdecl;
  external libc name PU + 'memcpy';
function _malloc(size: size_t): Pointer; cdecl;
  external libc name PU + 'malloc';
procedure _free(P: Pointer); cdecl; external libc name PU + 'free';
function _realloc(P: Pointer; NewSize: size_t): Pointer; cdecl;
  external libc name PU + 'realloc';
function _lseek(filedes: Integer; offset: off_t; whence: Integer): off_t; cdecl;
  external libc name PU + 'lseek';
function _strlen(S1: MarshaledAString): size_t; cdecl;
  external libc name PU + 'strlen';
function _strcat(S1: MarshaledAString; s2: MarshaledAString): MarshaledAString;
  cdecl; external libc name PU + 'strcat';
{$ENDIF}
{$IF defined(WIN32)}
{$IFDEF WINXP}
function CryptAcquireContextW(phProv: Pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: Pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
{$ELSE}
function BCryptOpenAlgorithmProvider(out phAlgorithm: THandle;
  pszAlgId: LPCWSTR; pszProvider: LPCWSTR; dwProvType: DWORD; dwFlags: DWORD)
  : LongInt; stdcall; external BCRYPT name 'BCryptOpenAlgorithmProvider';
function BCryptCloseAlgorithmProvider(hAlgorithm: THandle; dwFlags: ULONG)
  : LongInt; stdcall; external BCRYPT name 'BCryptCloseAlgorithmProvider';
function BCryptGenRandom(hAlgorithm: THandle; out pbBuffer: PUCHAR;
  cbBuffer, dwFlags: ULONG): LongInt; stdcall;
  external BCRYPT name 'BCryptGenRandom';
{$ENDIF}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomUBuffer';
function _fopen(Str1: PByte; Str2: PByte): Pointer; cdecl;
  external msvcrt name 'fopen';
function _fread(Ptr1: Pointer; Int1: Integer; Int2: Integer; Ptr2: Pointer)
  : Integer; cdecl; external msvcrt name 'fread';
function _fseek(Ptr1: Pointer; Int1: Integer; Int2: Integer): Integer; cdecl;
  external msvcrt name 'fseek';
function _ftell(Ptr: Pointer): Integer; cdecl; external msvcrt name 'ftell';
function _fwrite(Ptr1: Pointer; Int1: Integer; Int2: Integer; Ptr2: Pointer)
  : Integer; cdecl; external msvcrt name 'fwrite';
function _fclose(Ptr: Pointer): Integer; cdecl; external msvcrt name 'fclose';

{$ELSEIF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
function CryptAcquireContextW(phProv: Pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: Pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomBuffer';
function RandomInt(): Integer; external name PU + 'RandomInt';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomUBuffer';
function RandomUInt(): Integer; external name PU + 'RandomUInt';
{$ELSE}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external RANDOMDLL name 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external RANDOMDLL name 'RandomUBuffer';
function RandomInt(): Integer; external RANDOMDLL name 'RandomInt';
{$IFEND}
function fclose(Ptr: Pointer): Integer; external msvcrt name 'fclose';

function fopen(Str1: PAnsiChar; Str2: PAnsiChar): Pointer;
  external msvcrt name 'fopen';

function fread(Ptr1: Pointer; Int1: Integer; Int2: Integer; Ptr2: Pointer)
  : Integer; external msvcrt name 'fread';

function fseek(Ptr1: Pointer; Int1: Integer; Int2: Integer): Integer;
  external msvcrt name 'fseek';

function ftell(Ptr: Pointer): Integer; external msvcrt name 'ftell';

function fwrite(Ptr1: Pointer; Int1: Integer; Int2: Integer; Ptr2: Pointer)
  : Integer; external msvcrt name 'fwrite';
procedure __chkstk; external 'ntdll';
{$IFEND}
{$ELSEIF defined(ANDROID) or defined(IOS)}
procedure unsignedIntToTab(w: Integer; tab: PByte); external 'libTMSCPLib.a';
function tabToUnsignedLongLong(tab: PByte): LongInt; external 'libTMSCPLib.a';
procedure permutation_blake2b(v: PLongInt; vo: Integer; v1: Integer;
  v2: Integer; v3: Integer; v4: Integer; v5: Integer; v6: Integer; v7: Integer;
  v8: Integer; v9: Integer; v10: Integer; v11: Integer; v12: Integer;
  v13: Integer; v14: Integer; v15: Integer); external 'libTMSCPLib.a';
procedure unsignedLongLongToTab(w: LongInt; tab: PByte);
  external 'libTMSCPLib.a';

function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomInt(): Integer; external 'libTMSCPLib.a';
function SHA3_XOF_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PByte): Integer;
  external 'libTMSCPLib.a';
function SHA3_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer; external 'libTMSCPLib.a';
function SHA3_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer; external 'libTMSCPLib.a';
function SHA3_XOF_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  version: Integer; fileHash: PByte): Integer; external 'libTMSCPLib.a';
function SHA3_cSHAKE_BufferHash(inputBuffer: PByte; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64; bufferHash: PByte;
  version: Integer): Integer; external 'libTMSCPLib.a';
function SHA3_KMAC_BufferHash(key: PByte; keySizeBytes: Int64;
  inputBuffer: PByte; inputSizeBytes: Int64; hashSizeBits: Int64; salt: PByte;
  saltSizeBytes: Int64; bufferHash: PByte; version: Integer): Integer;
  external 'libTMSCPLib.a';
function SHA3_TupleHASH_BufferHash(inputBuffer: PPByte; inputSizeBytes: PInt64;
  nbInputBuffer: Int64; hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64;
  bufferHash: PByte; version: Integer): Integer; external 'libTMSCPLib.a';
procedure KeccakF1600_StatePermute(var state: TSHA3Array);
  external 'libTMSCPLib.a';
function Blake2B_FileHash(input_path: PByte; key: PByte; keylen: Integer;
  output: PByte; outputSizeBytes: Integer): Integer; external 'libTMSCPLib.a';
function Blake2B_BufferHash(input: PByte; inputSizeBytes: Integer; key: PByte;
  keylen: Integer; output: PByte; outputSizeBytes: Integer): Integer;
  external 'libTMSCPLib.a';
function update_blake2b(s: Pblake2b_state; input: PByte;
  inputSizeBytes: Integer): Integer; external 'libTMSCPLib.a';
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer): Integer;
  external 'libTMSCPLib.a';
function finalHash_blake2b(s: Pblake2b_state; outB: PByte;
  outputSizeBytes: Integer): Integer; external 'libTMSCPLib.a';
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PByte; keylen: Integer): Integer; external 'libTMSCPLib.a';
procedure RIPEMD160_BufferHash(input: PByte; inputSizeBytes: Integer;
  output: PByte); external 'libTMSCPLib.a';
function RIPEMD160_FileHash(inputPath: PByte; output: PByte): Integer;
  external 'libTMSCPLib.a';
procedure init_ripemd160(var buf: TRIPEMDBuf); external 'libTMSCPLib.a';
procedure compress_ripemd160(var buf: TRIPEMDBuf; X: TRIPEMDX);
  external 'libTMSCPLib.a';
procedure finish_ripemd160(var buf: TRIPEMDBuf; str: PByte; lswlen: Integer;
  mswlen: Integer); external 'libTMSCPLib.a';
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PByte; pwdSizeBytes: Integer; salt: saltArg; output: PByte;
  outputSizeBytes: Integer): Integer; external 'libTMSCPLib.a';
function SHA2_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer; external 'libTMSCPLib.a';
function SHA1_FileHash(inputFilePath: PByte; fileHash: PByte): Integer;
  external 'libTMSCPLib.a';
function SHA2_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer; external 'libTMSCPLib.a';
function SHA1_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  bufferHash: PByte): Integer; external 'libTMSCPLib.a';
procedure SHA256_BlockHash(mess: TSHA256Block; k: TSHA256Key; var h: TSHA256H);
  external 'libTMSCPLib.a';
procedure SHA384_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  external 'libTMSCPLib.a';
procedure SHA512_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  external 'libTMSCPLib.a';
procedure Constants_sha2(var k: TSHA256Key); external 'libTMSCPLib.a';
procedure Constants_sha512(var k: TSHA512Key); external 'libTMSCPLib.a';
function SHA2_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
  external 'libTMSCPLib.a';
function SHA2_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPLib.a';
function PBKDF2_KeyDerivationFunction_OLD(password: PByte; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PByte): Integer;
  external 'libTMSCPLib.a';
function SHA3_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
  external 'libTMSCPLib.a';
function SHA3_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPLib.a';
function SHA3_XOF_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; version: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPLib.a';
function SHA3_XOF_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; version: Integer;
  HMAC: PByte): Integer; external 'libTMSCPLib.a';
function PBKDF2_KeyDerivationFunction(password: PByte; inputSizeBytes: Integer;
  salt: PByte; saltSizeBytes: Integer; counter: Integer; h: Integer;
  hashSizeBits: Integer; outputSizeBits: Integer; output: PByte): Integer;
  external 'libTMSCPLib.a';
function CharToHexa(MyChar: PByte; LenChar: Integer; HexaChar: PByte): Integer;
  external 'libTMSCPLib.a';
function HexaToChar(HexaChar: PByte; MyChar: PByte): Integer;
  external 'libTMSCPLib.a';
{$ELSEIF defined(LINUX)}
procedure unsignedIntToTab(w: Integer; tab: PByte); external 'libTMSCPlib.a';
function tabToUnsignedLongLong(tab: PByte): LongInt; external 'libTMSCPlib.a';
procedure permutation_blake2b(v: PLongInt; vo: Integer; v1: Integer;
  v2: Integer; v3: Integer; v4: Integer; v5: Integer; v6: Integer; v7: Integer;
  v8: Integer; v9: Integer; v10: Integer; v11: Integer; v12: Integer;
  v13: Integer; v14: Integer; v15: Integer); external 'libTMSCPlib.a';
procedure unsignedLongLongToTab(w: LongInt; tab: PByte);
  external 'libTMSCPlib.a';

function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomInt(): Integer; external 'libTMSCPlib.a';
function SHA3_XOF_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PByte): Integer;
  external 'libTMSCPlib.a';
function SHA3_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer; external 'libTMSCPlib.a';
function SHA3_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer; external 'libTMSCPlib.a';
function SHA3_XOF_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  version: Integer; fileHash: PByte): Integer; external 'libTMSCPlib.a';
function SHA3_cSHAKE_BufferHash(inputBuffer: PByte; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64; bufferHash: PByte;
  version: Integer): Integer; external 'libTMSCPlib.a';
function SHA3_KMAC_BufferHash(key: PByte; keySizeBytes: Int64;
  inputBuffer: PByte; inputSizeBytes: Int64; hashSizeBits: Int64; salt: PByte;
  saltSizeBytes: Int64; bufferHash: PByte; version: Integer): Integer;
  external 'libTMSCPlib.a';
function SHA3_TupleHASH_BufferHash(inputBuffer: PPByte; inputSizeBytes: PInt64;
  nbInputBuffer: Int64; hashSizeBits: Int64; salt: PByte; saltSizeBytes: Int64;
  bufferHash: PByte; version: Integer): Integer; external 'libTMSCPlib.a';
procedure KeccakF1600_StatePermute(var state: TSHA3Array);
  external 'libTMSCPlib.a';
function Blake2B_FileHash(input_path: PByte; key: PByte; keylen: Integer;
  output: PByte; outputSizeBytes: Integer): Integer; external 'libTMSCPlib.a';
function Blake2B_BufferHash(input: PByte; inputSizeBytes: Integer; key: PByte;
  keylen: Integer; output: PByte; outputSizeBytes: Integer): Integer;
  external 'libTMSCPlib.a';
function update_blake2b(s: Pblake2b_state; input: PByte;
  inputSizeBytes: Integer): Integer; external 'libTMSCPlib.a';
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer): Integer;
  external 'libTMSCPlib.a';
function finalHash_blake2b(s: Pblake2b_state; outB: PByte;
  outputSizeBytes: Integer): Integer; external 'libTMSCPlib.a';
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PByte; keylen: Integer): Integer; external 'libTMSCPlib.a';
procedure RIPEMD160_BufferHash(input: PByte; inputSizeBytes: Integer;
  output: PByte); external 'libTMSCPlib.a';
function RIPEMD160_FileHash(inputPath: PByte; output: PByte): Integer;
  external 'libTMSCPlib.a';
procedure init_ripemd160(var buf: TRIPEMDBuf); external 'libTMSCPlib.a';
procedure compress_ripemd160(var buf: TRIPEMDBuf; X: TRIPEMDX);
  external 'libTMSCPlib.a';
procedure finish_ripemd160(var buf: TRIPEMDBuf; str: PByte; lswlen: Integer;
  mswlen: Integer); external 'libTMSCPlib.a';
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PByte; pwdSizeBytes: Integer; salt: saltArg; output: PByte;
  outputSizeBytes: Integer): Integer; external 'libTMSCPlib.a';
function SHA2_FileHash(inputFilePath: PByte; hashSizeBits: Integer;
  fileHash: PByte): Integer; external 'libTMSCPlib.a';
function SHA1_FileHash(inputFilePath: PByte; fileHash: PByte): Integer;
  external 'libTMSCPlib.a';
function SHA2_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PByte): Integer; external 'libTMSCPlib.a';
function SHA1_BufferHash(inputBuffer: PByte; inputSizeBytes: Integer;
  bufferHash: PByte): Integer; external 'libTMSCPlib.a';
procedure SHA256_BlockHash(mess: TSHA256Block; k: TSHA256Key; var h: TSHA256H);
  external 'libTMSCPlib.a';
procedure SHA384_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  external 'libTMSCPlib.a';
procedure SHA512_BlockHash(mess: TSHA512Block; k: TSHA512Key; var h: TSHA512H);
  external 'libTMSCPlib.a';
procedure Constants_sha2(var k: TSHA256Key); external 'libTMSCPlib.a';
procedure Constants_sha512(var k: TSHA512Key); external 'libTMSCPlib.a';
function SHA2_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
  external 'libTMSCPlib.a';
function SHA2_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPlib.a';
function PBKDF2_KeyDerivationFunction_OLD(password: PByte; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PByte): Integer;
  external 'libTMSCPlib.a';
function SHA3_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte): Integer;
  external 'libTMSCPlib.a';
function SHA3_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPlib.a';
function SHA3_XOF_HMAC(inputBuffer: PByte; inputSizeBytes: Integer; key: PByte;
  keySizeBytes: Integer; hashSizeBits: Integer; version: Integer; HMAC: PByte)
  : Integer; external 'libTMSCPlib.a';
function SHA3_XOF_VerifyHMAC(inputBuffer: PByte; inputSizeBytes: Integer;
  key: PByte; keySizeBytes: Integer; hashSizeBits: Integer; version: Integer;
  HMAC: PByte): Integer; external 'libTMSCPlib.a';
function PBKDF2_KeyDerivationFunction(password: PByte; inputSizeBytes: Integer;
  salt: PByte; saltSizeBytes: Integer; counter: Integer; h: Integer;
  hashSizeBits: Integer; outputSizeBits: Integer; output: PByte): Integer;
  external 'libTMSCPlib.a';
function CharToHexa(MyChar: PByte; LenChar: Integer; HexaChar: PByte): Integer;
  external 'libTMSCPlib.a';
function HexaToChar(HexaChar: PByte; MyChar: PByte): Integer;
  external 'libTMSCPlib.a';
{$ELSE}
function SHA3_XOF_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; version: Integer; bufferHash: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/sha3.o' name 'SHA3_XOF_BufferHash';
function SHA3_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_BufferHash';
function SHA3_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_FileHash';
function SHA3_XOF_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  version: Integer; fileHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_XOF_FileHash';
function SHA3_cSHAKE_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Int64;
  hashSizeBits: Int64; salt: PAnsiChar; saltSizeBytes: Int64;
  bufferHash: PAnsiChar; version: Integer): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_cSHAKE_BufferHash';
function SHA3_KMAC_BufferHash(key: PAnsiChar; keySizeBytes: Int64;
  inputBuffer: PAnsiChar; inputSizeBytes: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_KMAC_BufferHash';
function SHA3_TupleHASH_BufferHash(inputBuffer: PPAnsiChar;
  inputSizeBytes: PInt64; nbInputBuffer: Int64; hashSizeBits: Int64;
  salt: PAnsiChar; saltSizeBytes: Int64; bufferHash: PAnsiChar;
  version: Integer): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_TupleHASH_BufferHash';
procedure KeccakF1600_StatePermute(var state: TSHA3Array); cdecl;
  external object'./libOSX64/sha3.o' name 'KeccakF1600_StatePermute';
function Blake2B_FileHash(input_path: PAnsiChar; key: PAnsiChar;
  keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer): Integer; cdecl;
  external object'./libOSX64/blake2b.o' name 'Blake2B_FileHash';
function Blake2B_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer; output: PAnsiChar; outputSizeBytes: Integer)
  : Integer; cdecl;
  external object'./libOSX64/blake2b.o' name 'Blake2B_BufferHash';
function update_blake2b(s: Pblake2b_state; input: PAnsiChar;
  inputSizeBytes: Integer): Integer; cdecl;
  external object'./libOSX64/blake2b.o' name 'update_blake2b';
function init_blake2b(s: Pblake2b_state; outputSizeBytes: Integer): Integer;
  cdecl; external object'./libOSX64/blake2b.o' name 'init_blake2b';
function finalHash_blake2b(s: Pblake2b_state; outB: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
  external object'./libOSX64/blake2b.o' name 'finalHash_blake2b';
function initKey_blake2b(s: Pblake2b_state; outputSizeBytes: Integer;
  key: PAnsiChar; keylen: Integer): Integer; cdecl;
  external object'./libOSX64/blake2b.o' name 'initKey_blake2b';
procedure RIPEMD160_BufferHash(input: PAnsiChar; inputSizeBytes: Integer;
  output: PAnsiChar); cdecl;
  external object'./libOSX64/ripemd160.o' name 'RIPEMD160_BufferHash';
function RIPEMD160_FileHash(inputPath: PAnsiChar; output: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/ripemd160.o' name 'RIPEMD160_FileHash';
procedure init_ripemd160(buf: PInteger); cdecl;
  external object'./libOSX64/ripemd160.o' name 'init_ripemd160';
procedure compress_ripemd160(buf: PInteger; X: PInteger); cdecl;
  external object'./libOSX64/ripemd160.o' name 'compress_ripemd160';
procedure finish_ripemd160(buf: PInteger; str: PAnsiChar; lswlen: Integer;
  mswlen: Integer); cdecl;
  external object'./libOSX64/ripemd160.o' name 'finish_ripemd160';
function Argon2_KeyDerivation(counter: Integer; memory: Integer;
  password: PAnsiChar; pwdSizeBytes: Integer; salt: saltArg; output: PAnsiChar;
  outputSizeBytes: Integer): Integer; cdecl;
  external object'./libOSX64/argon2d.o' name 'Argon2_KeyDerivation';
function SHA2_FileHash(inputFilePath: PAnsiChar; hashSizeBits: Integer;
  fileHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha2.o' name 'SHA2_FileHash';
function SHA1_FileHash(inputFilePath: PAnsiChar; fileHash: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/sha2.o' name 'SHA1_FileHash';
function SHA2_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  hashSizeBits: Integer; bufferHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha2.o' name 'SHA2_BufferHash';
function SHA1_BufferHash(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  bufferHash: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha2.o' name 'SHA1_BufferHash';
procedure SHA256_BlockHash(mess: PAnsiChar; k: PCardinal; h: PCardinal); cdecl;
  external object'./libOSX64/sha2.o' name 'SHA256_BlockHash';
procedure SHA512_BlockHash(mess: PAnsiChar; k: PInt64; h: PInt64); cdecl;
  external object'./libOSX64/sha2.o' name 'SHA512_BlockHash';
procedure SHA384_BlockHash(mess: PAnsiChar; k: PInt64; h: PInt64); cdecl;
  external object'./libOSX64/sha2.o' name 'SHA384_BlockHash';
procedure Constants_sha2(var k: TSHA256Key); cdecl;
  external object'./libOSX64/sha2.o' name 'Constants_sha2';
procedure Constants_sha512(var k: TSHA512Key); cdecl;
  external object'./libOSX64/sha2.o' name 'Constants_sha512';
function SHA2_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/sha2.o' name 'SHA2_HMAC';
function SHA2_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/sha2.o' name 'SHA2_VerifyHMAC';
function PBKDF2_KeyDerivationFunction_OLD(password: PAnsiChar; salt: saltPB;
  counter: Integer; outputSizeBits: Integer; output: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/pbkdf2_old.o' name
  'PBKDF2_KeyDerivationFunction_OLD';
function SHA3_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/sha3.o' name 'SHA3_HMAC';
function SHA3_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer; HMAC: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/sha3.o' name 'SHA3_VerifyHMAC';
function SHA3_XOF_HMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_XOF_HMAC';
function SHA3_XOF_VerifyHMAC(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  key: PAnsiChar; keySizeBytes: Integer; hashSizeBits: Integer;
  version: Integer; HMAC: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/sha3.o' name 'SHA3_XOF_VerifyHMAC';
function PBKDF2_KeyDerivationFunction(password: PAnsiChar;
  inputSizeBytes: Integer; salt: PAnsiChar; saltSizeBytes: Integer;
  counter: Integer; h: Integer; hashSizeBits: Integer; outputSizeBits: Integer;
  output: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/pbkdf2.o' name 'PBKDF2_KeyDerivationFunction';
function CharToHexa(MyChar: PAnsiChar; LenChar: Integer; HexaChar: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/elt_rep.o' name 'CharToHexa';
function HexaToChar(HexaChar: PAnsiChar; MyChar: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'HexaToChar';
{$IFEND}

constructor TSHA2Hash.Create(AOwner: TComponent);
begin
  self.SetHashSizeBits(256);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
  inherited Create(AOwner);
end;

constructor TSHA2Hash.Create;
begin
  self.SetHashSizeBits(256);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
end;

constructor TSHA2Hash.Create(hashSizeBits: Integer; outputFormat: TConvertType;
  uni: TUnicode);
begin
  self.SetHashSizeBits(hashSizeBits);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
  self.FOnChange := nil;
end;

function TSHA2Hash.Hash(s: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash32: array [0 .. 31] of AnsiChar;
  bufferHash64: array [0 .. 63] of AnsiChar;
  bufferHash48: array [0 .. 47] of AnsiChar;
  sPB: PAnsiChar;
{$ELSE}
  bufferHash32, bufferHash64, bufferHash48, sPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create;
  conv.AType := outputFormat;
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  try
    case FHashSizeBits of
      256:
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          FillChar(bufferHash32, 32, #0);
{$ELSE}
          bufferHash32 := GetMemory(32 * SizeOf(Byte));
          FillChar(bufferHash32^, 32, Byte(#0));
{$IFEND}
          try
            err := SHA2_BufferHash(sPB, msgLen, 256, bufferHash32);

            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash32, 32);
          finally
{$IF not((defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS)))}
            FreeMem(bufferHash32);
{$IFEND}
          end;

        end;
      384:
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          FillChar(bufferHash48, 48, #0);
{$ELSE}
          bufferHash48 := GetMemory(48 * SizeOf(Byte));
          FillChar(bufferHash48^, 48, Byte(#0));
{$IFEND}
          try
            err := SHA2_BufferHash(sPB, msgLen, 384, bufferHash48);

            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash48, 48);
          finally
{$IF not((defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS)))}
            FreeMem(bufferHash48);
{$IFEND}
          end;

        end;
      512:
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          FillChar(bufferHash64, 64, #0);
{$ELSE}
          bufferHash64 := GetMemory(64 * SizeOf(Byte));
          FillChar(bufferHash64^, 64, #0);
{$IFEND}
          try
            err := SHA2_BufferHash(sPB, msgLen, 512, bufferHash64);

            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash64, 64);
          finally
{$IF not((defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS)))}
            FreeMem(bufferHash64);
{$IFEND}
          end;

        end;
    end;
  finally
    conv.Free;
    FreeMem(sPB);
  end;

end;

function TSHA2Hash.HashFile(s: string): string;
var
  stream1: TFileStream;
begin

  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    Result := self.HashStream(stream1);
  finally
    stream1.Free;
  end;

end;

function TSHA2Hash.HashStream(s: TStream): string;
type
  uint64 = record
    case boolean of
      true:
        (i64: Int64);
      false:
        (lo, hi: Cardinal);
  end;

var
{$IF defined(MACOS64) and not defined(IOS)}
  h256: PCardinal;
  h512, h384: PInt64;
  k256: PCardinal;
  k512, k384: PInt64;
  mess256: PAnsiChar;
  mess512, mess384: PAnsiChar;
  tempk256: TSHA256Key;
  tempk512, tempk384: TSHA512Key;
  tempmess256: TSHA256Block;
  tempmess384, tempmess512: TSHA512Block;
{$ELSE}
  h256: TSHA256H;
  h512, h384: TSHA512H;
  k256: TSHA256Key;
  k512, k384: TSHA512Key;
  mess256: TSHA256Block;
  mess512, mess384: TSHA512Block;
{$IFEND}
  err, I, j, N: Integer;
  len_m, l, nbBlocksRead: Int64;
  padd_prev_block: boolean;
  conv: TConvert;
  k64: uint64;
begin
  conv := TConvert.Create(self.outputFormat);
  try
    s.Position := 0;
    SetProgress(0);
    l := s.size * 8;
    padd_prev_block := false;
    if self.FHashSizeBits = 256 then
    begin
{$IF defined(MACOS64) and not defined(IOS)}
      h256 := GetMemory(8 * SizeOf(Cardinal));
      k256 := GetMemory(64 * SizeOf(Cardinal));
      mess256 := GetMemory(64 * SizeOf(AnsiChar));
      try
{$IFEND}
        h256[0] := $6A09E667;
        h256[1] := $BB67AE85;
        h256[2] := $3C6EF372;
        h256[3] := $A54FF53A;
        h256[4] := $510E527F;
        h256[5] := $9B05688C;
        h256[6] := $1F83D9AB;
        h256[7] := $5BE0CD19;

        // padding
        len_m := l + 1 + 64;

        while len_m mod 512 <> 0 do
          len_m := len_m + 1;

        N := len_m div 512;

{$IF defined(MACOS64) and not defined(IOS)}
        Constants_sha2(tempk256);
        for I := 0 to 63 do
          k256[I] := tempk256[I];
{$ELSE}
        Constants_sha2(k256);
{$IFEND}
        nbBlocksRead := 0;
        for I := 0 to N - 2 do
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          FillChar(tempmess256, 64, 0);
{$ELSE}
          FillChar(mess256, 64, 0);
{$IFEND}
          padd_prev_block := false;
{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess256[0], 64 * SizeOf(tempmess256[0]));
          for j := 0 to 63 do
            mess256[j] := tempmess256[j];
{$ELSE}
          err := s.Read(mess256[0], 64 * SizeOf(mess256[0]));
{$IFEND}
          nbBlocksRead := nbBlocksRead + 64;
          if err <> 64 then
          begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
            mess256[err] := AnsiChar(128);
{$ELSE}
            mess256[err] := Byte(128);
{$IFEND}
            padd_prev_block := true;
          end;
          SHA256_BlockHash(mess256, k256, h256);
          SetProgress((nbBlocksRead * 100) div s.size);
        end;

{$IF defined(MACOS64) and not defined(IOS)}
        FillChar(tempmess256, 64, 0);
        FillChar(mess256^, 64, 0);
{$ELSE}
        FillChar(mess256, 64, 0);
{$IFEND}
        if padd_prev_block = false then
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess256[0], 64 * SizeOf(tempmess256[0]));
          for j := 0 to 63 do
            mess256[j] := tempmess256[j];
{$ELSE}
          err := s.Read(mess256[0], 64 * SizeOf(mess256[0]));
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess256[err] := AnsiChar(128);
{$ELSE}
          mess256[err] := Byte(128);
{$IFEND}
        end;

        for j := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess256[64 - (4 - j)] := AnsiChar(l shr (24 - (j * 8)) and 255);
{$ELSE}
          mess256[64 - (4 - j)] := Byte(l shr (24 - (j * 8)) and 255);
{$IFEND}
        nbBlocksRead := nbBlocksRead + 64;
        SHA256_BlockHash(mess256, k256, h256);
        SetProgress((nbBlocksRead * 100) div s.size);

        Result := '';
        for I := 0 to 7 do
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          Result := Result + string(AnsiChar(h256[I] shr 24));
          Result := Result + string(AnsiChar((h256[I] shr 16) and 255));
          Result := Result + string(AnsiChar((h256[I] shr 8) and 255));
          Result := Result + string(AnsiChar(h256[I] and 255));
{$ELSE}
          Result := Result + Char(h256[I] shr 24);
          Result := Result + Char((h256[I] shr 16) and 255);
          Result := Result + Char((h256[I] shr 8) and 255);
          Result := Result + Char(h256[I] and 255);
{$IFEND}
        end;

        Result := conv.StringToFormat(Result);
        SetProgress(100);
{$IF defined(MACOS64) and not defined(IOS)}
      finally
        FreeMem(h256);
        FreeMem(k256);
        FreeMem(mess256);
      end;
{$IFEND}
    end
    else if self.hashSizeBits = 384 then
    begin
{$IF defined(MACOS64) and not defined(IOS)}
      h384 := GetMemory(8 * SizeOf(Int64));
      k384 := GetMemory(80 * SizeOf(Int64));
      mess384 := GetMemory(128 * SizeOf(AnsiChar));
      try
{$IFEND}
        k64.hi := $CBBB9D5D;
        k64.lo := $C1059ED8;

        h384[0] := k64.i64;

        k64.hi := $629A292A;
        k64.lo := $367CD507;

        h384[1] := k64.i64;

        k64.hi := $9159015A;
        k64.lo := $3070DD17;

        h384[2] := k64.i64;

        k64.hi := $152FECD8;
        k64.lo := $F70E5939;

        h384[3] := k64.i64;

        k64.hi := $67332667;
        k64.lo := $FFC00B31;

        h384[4] := k64.i64;

        k64.hi := $8EB44A87;
        k64.lo := $68581511;

        h384[5] := k64.i64;

        k64.hi := $DB0C2E0D;
        k64.lo := $64F98FA7;

        h384[6] := k64.i64;

        k64.hi := $47B5481D;
        k64.lo := $BEFA4FA4;

        h384[7] := k64.i64;

        // padding
        len_m := l + 1 + 128;

        while len_m mod 1024 <> 0 do
          len_m := len_m + 1;

        N := len_m div 1024;

{$IF defined(MACOS64) and not defined(IOS)}
        Constants_sha512(tempk384);
        for I := 0 to 79 do
          k384[I] := tempk384[I];
{$ELSE}
        Constants_sha512(k384);
{$IFEND}
        nbBlocksRead := 0;
        for I := 0 to N - 2 do
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          FillChar(tempmess384, 128, 0);
{$ELSE}
          FillChar(mess384, 128, 0);
{$IFEND}
          padd_prev_block := false;

{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess384[0], 128 * SizeOf(tempmess384[0]));
          for j := 0 to 127 do
            mess384[j] := tempmess384[j];
{$ELSE}
          err := s.Read(mess384[0], 128 * SizeOf(mess384[0]));
{$IFEND}
          nbBlocksRead := nbBlocksRead + 64;
          if err <> 128 then
          begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
            mess384[err] := AnsiChar(128);
{$ELSE}
            mess384[err] := Byte(128);
{$IFEND}
            padd_prev_block := true;
          end;
          SHA384_BlockHash(mess384, k384, h384);
          SetProgress((nbBlocksRead * 100) div s.size);
        end;

{$IF defined(MACOS64) and not defined(IOS)}
        FillChar(tempmess384, 128, 0);
{$ELSE}
        FillChar(mess384, 128, 0);
{$IFEND}
        if padd_prev_block = false then
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess384[0], 128 * SizeOf(tempmess384[0]));
          for j := 0 to 127 do
            mess384[j] := tempmess384[j];
{$ELSE}
          err := s.Read(mess384[0], 128 * SizeOf(mess384[0]));
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess384[err] := AnsiChar(128);
{$ELSE}
          mess384[err] := Byte(128);
{$IFEND}
        end;
        nbBlocksRead := nbBlocksRead + 64;
        for j := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess384[128 - (4 - j)] := AnsiChar((l shr (24 - j * 8)) and 255);
{$ELSE}
          mess384[128 - (4 - j)] := Byte((l shr (24 - j * 8)) and 255);
{$IFEND}
        SHA384_BlockHash(mess384, k384, h384);
        SetProgress((nbBlocksRead * 100) div s.size);
        Result := '';

        for I := 0 to 5 do
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          Result := Result + string(AnsiChar(h384[I] shr 56));
          Result := Result + string(AnsiChar((h384[I] shr 48) and 255));
          Result := Result + string(AnsiChar((h384[I] shr 40) and 255));
          Result := Result + string(AnsiChar((h384[I] shr 32) and 255));
          Result := Result + string(AnsiChar((h384[I] shr 24) and 255));
          Result := Result + string(AnsiChar((h384[I] shr 16) and 255));
          Result := Result + string(AnsiChar((h384[I] shr 8) and 255));
          Result := Result + string(AnsiChar(h384[I] and 255));
{$ELSE}
          Result := Result + Char(h384[I] shr 56);
          Result := Result + Char((h384[I] shr 48) and 255);
          Result := Result + Char((h384[I] shr 40) and 255);
          Result := Result + Char((h384[I] shr 32) and 255);
          Result := Result + Char((h384[I] shr 24) and 255);
          Result := Result + Char((h384[I] shr 16) and 255);
          Result := Result + Char((h384[I] shr 8) and 255);
          Result := Result + Char(h384[I] and 255);
{$IFEND}
        end;
        SetProgress(100);
        Result := conv.StringToFormat(Result);
{$IF defined(MACOS64) and not defined(IOS)}
      finally
        FreeMem(h384);
        FreeMem(k384);
      end;
{$IFEND}
    end
    else
    begin
{$IF defined(MACOS64) and not defined(IOS)}
      h512 := GetMemory(8 * SizeOf(Int64));
      k512 := GetMemory(80 * SizeOf(Int64));
      mess512 := GetMemory(128 * SizeOf(AnsiChar));
      try
{$IFEND}
        k64.hi := $6A09E667;
        k64.lo := $F3BCC908;

        h512[0] := k64.i64;

        k64.hi := $BB67AE85;
        k64.lo := $84CAA73B;

        h512[1] := k64.i64;

        k64.hi := $3C6EF372;
        k64.lo := $FE94F82B;

        h512[2] := k64.i64;

        k64.hi := $A54FF53A;
        k64.lo := $5F1D36F1;

        h512[3] := k64.i64;

        k64.hi := $510E527F;
        k64.lo := $ADE682D1;

        h512[4] := k64.i64;

        k64.hi := $9B05688C;
        k64.lo := $2B3E6C1F;

        h512[5] := k64.i64;

        k64.hi := $1F83D9AB;
        k64.lo := $FB41BD6B;

        h512[6] := k64.i64;

        k64.hi := $5BE0CD19;
        k64.lo := $137E2179;

        h512[7] := k64.i64;

        // padding
        len_m := l + 1 + 128;

        while len_m mod 1024 <> 0 do
          len_m := len_m + 1;

        N := len_m div 1024;

{$IF defined(MACOS64) and not defined(IOS)}
        Constants_sha512(tempk512);
        for I := 0 to 79 do
          k512[I] := tempk512[I];
{$ELSE}
        Constants_sha512(k512);
{$IFEND}
        nbBlocksRead := 0;
        for I := 0 to N - 2 do
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          FillChar(tempmess512, 128, 0);
{$ELSE}
          FillChar(mess512, 128, 0);
{$IFEND}
          padd_prev_block := false;
{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess512[0], 128 * SizeOf(tempmess512[0]));
          for j := 0 to 127 do
            mess512[j] := tempmess512[j];
{$ELSE}
          err := s.Read(mess512[0], 128 * SizeOf(mess512[0]));
{$IFEND}
          nbBlocksRead := nbBlocksRead + 64;
          if err <> 128 then
          begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
            mess512[err] := AnsiChar(128);
{$ELSE}
            mess512[err] := Byte(128);
{$IFEND}
            padd_prev_block := true;
          end;
          SHA512_BlockHash(mess512, k512, h512);
          SetProgress((nbBlocksRead * 100) div s.size);
        end;

{$IF defined(MACOS64) and not defined(IOS)}
        FillChar(tempmess512, 128, 0);
{$ELSE}
        FillChar(mess512, 128, 0);
{$IFEND}
        if padd_prev_block = false then
        begin
{$IF defined(MACOS64) and not defined(IOS)}
          err := s.Read(tempmess512[0], 128 * SizeOf(tempmess512[0]));
          for j := 0 to 127 do
            mess512[j] := tempmess512[j];
{$ELSE}
          err := s.Read(mess512[0], 128 * SizeOf(mess512[0]));
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess512[err] := AnsiChar(128);
{$ELSE}
          mess512[err] := Byte(128);
{$IFEND}
        end;
        nbBlocksRead := nbBlocksRead + 64;
        for j := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          mess512[128 - (4 - j)] := AnsiChar((l shr (24 - j * 8)) and 255);
{$ELSE}
          mess512[128 - (4 - j)] := Byte((l shr (24 - j * 8)) and 255);
{$IFEND}
        SHA512_BlockHash(mess512, k512, h512);
        SetProgress((nbBlocksRead * 100) div s.size);
        Result := '';

        for I := 0 to 7 do
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          Result := Result + string(AnsiChar(h512[I] shr 56));
          Result := Result + string(AnsiChar((h512[I] shr 48) and 255));
          Result := Result + string(AnsiChar((h512[I] shr 40) and 255));
          Result := Result + string(AnsiChar((h512[I] shr 32) and 255));
          Result := Result + string(AnsiChar((h512[I] shr 24) and 255));
          Result := Result + string(AnsiChar((h512[I] shr 16) and 255));
          Result := Result + string(AnsiChar((h512[I] shr 8) and 255));
          Result := Result + string(AnsiChar(h512[I] and 255));
{$ELSE}
          Result := Result + Char(h512[I] shr 56);
          Result := Result + Char((h512[I] shr 48) and 255);
          Result := Result + Char((h512[I] shr 40) and 255);
          Result := Result + Char((h512[I] shr 32) and 255);
          Result := Result + Char((h512[I] shr 24) and 255);
          Result := Result + Char((h512[I] shr 16) and 255);
          Result := Result + Char((h512[I] shr 8) and 255);
          Result := Result + Char(h512[I] and 255);
{$IFEND}
        end;
        SetProgress(100);
        Result := conv.StringToFormat(Result);
{$IF defined(MACOS64) and not defined(IOS)}
      finally
        FreeMem(h512);
        FreeMem(k512);
        FreeMem(mess512);
      end;
{$IFEND}
    end;
  finally
    conv.Free;
  end;

end;

function TSHA2Hash.HMAC(s, k: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  HMAC, sPB, kPB: PAnsiChar;
{$ELSE}
  HMAC, sPB, kPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  kPB := conv.StringToBufferA(k, noUni);
  try
    HMAC := GetMemory(FHashSizeBits div 8);
    FillChar(HMAC^, FHashSizeBits div 8, #0);
    try
      err := SHA2_HMAC(sPB, msgLen, kPB, Length(k), FHashSizeBits, HMAC);

      if err <> 0 then
      begin
        raise ECryptoPack.Create(err);
      end;
      Result := conv.CharToFormat(HMAC, FHashSizeBits div 8);
    finally
      FreeMem(HMAC);
    end;

  finally
    conv.Free;
    FreeMem(sPB);
    FreeMem(kPB);
  end;

end;

function TSHA2Hash.VerifyHMAC(s, k, h: string): Integer;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  sPB, kPB, HMAC: PAnsiChar;
{$ELSE}
  sPB, kPB, HMAC: PByte;
{$IFEND}
  conv: TConvert;
begin
  err := 0;
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  kPB := conv.StringToBufferA(k, noUni);
  try

    begin
      HMAC := conv.FormatToChar(h);
      try
        err := SHA2_VerifyHMAC(sPB, msgLen, kPB, Length(k),
          FHashSizeBits, HMAC);

        if (err <> 0) and (err <> -106) then
        begin
          raise ECryptoPack.Create(err);
        end;
      finally
        FreeMem(HMAC);
      end;

    end;

  finally
    Result := err;
    conv.Free;
    FreeMem(kPB);
    FreeMem(sPB);
  end;

end;

procedure TSHA2Hash.SetHashSizeBits(const Value: Integer);
begin
  if (Value <> 256) and (Value <> 512) and (Value <> 384) then
    raise ECryptoPack.Create
      ('Incorrect size of the hash, must be 256, 384 or 512');
  FHashSizeBits := Value;
end;

procedure TSHA2Hash.SetProgress(P: Integer);
begin
  if P <> FProgress then
  begin
    FProgress := P;
    if Assigned(FOnChange) then
      FOnChange(self);
  end;
end;

Constructor TSHA3Hash.Create(AOwner: TComponent);
begin
  self.SetType(tsha);
  self.SetHashSizeBits(256);
  self.SetVersion(256);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
  inherited Create(AOwner);
end;

Constructor TSHA3Hash.Create;
begin
  self.SetType(tsha);
  self.SetHashSizeBits(256);
  self.SetVersion(256);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
end;

Constructor TSHA3Hash.Create(hashSizeBits: Integer; outputFormat: TConvertType;
  uni: TUnicode);
begin
  self.SetType(tsha);
  self.SetHashSizeBits(hashSizeBits);
  self.SetVersion(256);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
  self.FOnChange := nil;
end;

Constructor TSHA3Hash.Create(hashSizeBits: Integer; outputFormat: TConvertType;
  uni: TUnicode; version: Integer);
begin
  self.SetType(txof);
  self.SetVersion(version);
  self.SetHashSizeBits(hashSizeBits);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
  self.FOnChange := nil;
end;

function TSHA3Hash.Hash(s: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, sPB: PAnsiChar;
{$ELSE}
  bufferHash, sPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  try
    case FType of
      tsha:
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          bufferHash := GetMemory((FHashSizeBits DIV 8) * SizeOf(AnsiChar));
{$ELSE}
          bufferHash := GetMemory((FHashSizeBits DIV 8) * SizeOf(Byte));
{$IFEND}
          try
            err := SHA3_BufferHash(sPB, msgLen, FHashSizeBits, bufferHash);
            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash, FHashSizeBits DIV 8);
          finally
            FreeMem(bufferHash);
          end;

        end;
      txof:
        begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
          bufferHash := GetMemory((hashSizeBits DIV 8) * SizeOf(AnsiChar));
{$ELSE}
          bufferHash := GetMemory((hashSizeBits DIV 8) * SizeOf(Byte));
{$IFEND}
          try
            err := SHA3_XOF_BufferHash(sPB, msgLen, FHashSizeBits, FVersion,
              bufferHash);
            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash, FHashSizeBits DIV 8);
          finally
            FreeMem(bufferHash);
          end;

        end;
    end;
  finally
    FreeMem(sPB);
    conv.Free;
  end;

end;

function TSHA3Hash.cSHAKEHash(s, salt: string): string;
var
  err, msgLen, saltLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, sPB, saltPB: PAnsiChar;
{$ELSE}
  bufferHash, sPB, saltPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  saltPB := conv.StringToBuffer(salt, FUni, saltLen);
  try
    case FType of
      tsha:
        raise ECryptoPack.Create('This function can only be used in txof mode');
      txof:
        begin
          bufferHash := GetMemory(FHashSizeBits DIV 8);
          try
            err := SHA3_cSHAKE_BufferHash(sPB, msgLen, FHashSizeBits, saltPB,
              saltLen, bufferHash, FVersion);
            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash, FHashSizeBits DIV 8);
          finally
            FreeMem(bufferHash);
          end;

        end;
    end;
  finally
    FreeMem(sPB);
    FreeMem(saltPB);
    conv.Free;
  end;

end;

function TSHA3Hash.KMACHash(k, s, salt: string): string;
var
  err, msgLen, saltLen, kLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, sPB, saltPB, kPB: PAnsiChar;
{$ELSE}
  bufferHash, sPB, saltPB, kPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  saltPB := conv.StringToBuffer(salt, FUni, saltLen);
  kPB := conv.StringToBuffer(k, FUni, kLen);
  try
    case FType of
      tsha:
        raise ECryptoPack.Create('This function can only be used in txof mode');
      txof:
        begin
          bufferHash := GetMemory(hashSizeBits DIV 8);
          try
            err := SHA3_KMAC_BufferHash(kPB, kLen, sPB, msgLen, FHashSizeBits,
              saltPB, saltLen, bufferHash, FVersion);

            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash, FHashSizeBits DIV 8);
          finally
            FreeMem(bufferHash);
          end;

        end;
    end;
  finally
    FreeMem(sPB);
    FreeMem(saltPB);
    FreeMem(kPB);
    conv.Free;
  end;

end;

function TSHA3Hash.TupleHash(s: array of string; salt: string): string;
var
  err, msgLen, saltLen, temp: Integer;
  sLen: PInt64;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, saltPB: PAnsiChar;
  sPB: PPAnsiChar;
{$ELSE}
  bufferHash, saltPB: PByte;
  sPB: PPByte;
{$IFEND}
  conv: TConvert;
  I: Integer;
begin
  conv := TConvert.Create(outputFormat);
  msgLen := Length(s);
  sLen := GetMemory(msgLen * SizeOf(Int64));
  sPB := GetMemory(msgLen);
  for I := Low(s) to High(s) do
  begin
    sPB[I] := conv.StringToBuffer(s[I], FUni, temp);
    sLen[I] := temp;
  end;

  saltPB := conv.StringToBuffer(salt, FUni, saltLen);
  try
    case FType of
      tsha:
        raise ECryptoPack.Create('This function can only be used in txof mode');
      txof:
        begin
          bufferHash := GetMemory(hashSizeBits DIV 8);
          try
            err := SHA3_TupleHASH_BufferHash(sPB, sLen, msgLen, FHashSizeBits,
              saltPB, saltLen, bufferHash, FVersion);

            if err <> 0 then
            begin
              raise ECryptoPack.Create(err);
            end;
            Result := conv.CharToFormat(bufferHash, FHashSizeBits DIV 8);
          finally
            FreeMem(bufferHash);
          end;

        end;
    end;
  finally
    conv.Free;
    FreeMem(sLen);
    FreeMem(sPB);
    FreeMem(saltPB);
  end;

end;

function TSHA3Hash.HashFile(s: string): string;
var
  stream1: TFileStream;
begin
  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    Result := self.HashStream(stream1);
  finally
    stream1.Free;
  end;

end;

function TSHA3Hash.HashStream(s: TStream): string;
var
  state: array [0 .. 199] of Byte;
  rate, rateInBytes, blockSize, bRead, inputByteLen, I, capacity, index,
    outputByteLen: Integer;
  str: string;
  delimitedSuffix: Byte;
  nbBlocksRead: Int64;
  iBlock: array [0 .. 31823] of Byte;
  conv: TConvert;
begin
  conv := TConvert.Create(self.FOutputFormat);
  try
    s.Position := 0;
    blockSize := 0;
    SetProgress(0);
    outputByteLen := self.hashSizeBits div 8;

    if self.AType = tsha then
    begin
      delimitedSuffix := $06;
      capacity := 2 * self.hashSizeBits;
    end
    else
    begin
      delimitedSuffix := $1F;
      capacity := self.version;
    end;

    rate := 1600 - capacity;
    if rate mod 8 <> 0 then
      raise ECryptoPack.Create(-1006);
    rateInBytes := rate div 8;
    FillChar(state, SizeOf(state), 0);
    bRead := 1;
    FillChar(iBlock, SizeOf(iBlock), 0);
    nbBlocksRead := 0;
    while bRead <> 0 do
    begin
      bRead := s.Read(iBlock[0], min(s.size, 31824) * SizeOf(iBlock[0]));
      index := 0;
      inputByteLen := bRead;
      nbBlocksRead := nbBlocksRead + bRead;
      while inputByteLen > 0 do
      begin
        blockSize := min(inputByteLen, rateInBytes);
        for I := 0 to blockSize - 1 do
          state[I] := state[I] xor Byte(iBlock[I + index]);
        index := index + blockSize;
        inputByteLen := inputByteLen - blockSize;
        if blockSize = rateInBytes then
        begin
          KeccakF1600_StatePermute(TSHA3Array(state));
          blockSize := 0;
        end;
      end;
      SetProgress((nbBlocksRead * 100) div s.size);
    end;
    state[blockSize] := state[blockSize] xor delimitedSuffix;
    if (delimitedSuffix and $80 <> 0) and (blockSize = rateInBytes - 1) then
      KeccakF1600_StatePermute(TSHA3Array(state));
    state[rateInBytes - 1] := state[rateInBytes - 1] xor $80;
    KeccakF1600_StatePermute(TSHA3Array(state));
    Result := '';
    while outputByteLen > 0 do
    begin
      blockSize := min(outputByteLen, rateInBytes);
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      SetString(str, PAnsiChar(@state[0]), blockSize);
{$ELSE}
      str := '';
      for I := 0 to blockSize - 1 do
        str := str + Char(state[I]);
{$IFEND}
      Result := Result + str;
      outputByteLen := outputByteLen - blockSize;
      if outputByteLen > 0 then
        KeccakF1600_StatePermute(TSHA3Array(state));
    end;
    SetProgress(100);
    Result := conv.StringToFormat(Result);
  finally
    conv.Free;
  end;

end;

function TSHA3Hash.HMAC(s, k: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  HMAC, sPB, kPB: PAnsiChar;
{$ELSE}
  HMAC, sPB, kPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  kPB := conv.StringToBufferA(k, noUni);
  try

    begin
      HMAC := GetMemory(FHashSizeBits div 8);
      FillChar(HMAC^, (FHashSizeBits div 8), #0);
      try
        case AType of
          tsha:
            err := SHA3_HMAC(sPB, msgLen, kPB, Length(k), FHashSizeBits, HMAC);
        else
          err := SHA3_XOF_HMAC(sPB, msgLen, kPB, Length(k), FHashSizeBits,
            FVersion, HMAC);
        end;
        if err <> 0 then
          raise ECryptoPack.Create(err);
        Result := conv.CharToFormat(HMAC, (FHashSizeBits div 8));
      finally
        FreeMem(HMAC);
      end;
    end;
  finally
    conv.Free;
    FreeMem(sPB);
    FreeMem(kPB);
  end;

end;

function TSHA3Hash.VerifyHMAC(s, k, h: string): Integer;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  sPB, kPB, HMAC: PAnsiChar;
{$ELSE}
  sPB, kPB, HMAC: PByte;
{$IFEND}
  conv: TConvert;
begin
  err := 0;
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  kPB := conv.StringToBufferA(k, noUni);
  try
    HMAC := conv.FormatToChar(h);
    try
      case AType of
        tsha:
          err := SHA3_VerifyHMAC(sPB, msgLen, kPB, Length(k),
            FHashSizeBits, HMAC);
        txof:
          err := SHA3_XOF_VerifyHMAC(sPB, msgLen, kPB, Length(k), FHashSizeBits,
            FVersion, HMAC);
      end;
      if (err <> 0) and (err <> -1007) then
        raise ECryptoPack.Create(err);
    finally
      FreeMem(HMAC);
    end;

  finally
    Result := err;
    conv.Free;
    FreeMem(sPB);
    FreeMem(kPB);
  end;

end;

procedure TSHA3Hash.SetHashSizeBits(const Value: Integer);
begin
  case FType of
    tsha:
      begin
        if (Value <> 256) and (Value <> 384) and (Value <> 224) and
          (Value <> 512) then
          raise ECryptoPack.Create
            ('The size of the hash is incorrect, must be 224, 256, 384 or 512');
        FHashSizeBits := Value;
      end;
    txof:
      begin
        if Value mod 8 <> 0 then
          raise ECryptoPack.Create
            ('The SHA 3 XOF output length must be a multiple of 8!');
        FHashSizeBits := Value;
      end;
  end;
end;

procedure TSHA3Hash.SetProgress(P: Integer);
begin
  if P <> FProgress then
  begin
    FProgress := P;
    if Assigned(FOnChange) then
      FOnChange(self);
  end;
end;

procedure TSHA3Hash.SetType(const Value: TSHA3Type);
begin
  FType := Value;
end;

procedure TSHA3Hash.SetVersion(const Value: Integer);
begin
  if (Value <> 256) and (Value <> 512) then
    raise ECryptoPack.Create('The version is incorrect, must be 256 or 512');
  FVersion := Value;
end;

Constructor TPBKDF2KeyDerivationOLD.Create(AOwner: TComponent);
begin
  self.SetOutputSizeBits(128);
  self.SetCounter(10000);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  inherited Create(AOwner);
end;

Constructor TPBKDF2KeyDerivationOLD.Create;
begin
  self.SetOutputSizeBits(128);
  self.SetCounter(10000);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
end;

Constructor TPBKDF2KeyDerivationOLD.Create(outputSizeBits: Integer;
  salt: string; counter: Integer; outputFormat: TConvertType; uni: TUnicode);
begin
  self.SetOutputSizeBits(outputSizeBits);
  self.SetStringSalt(salt);
  self.SetCounter(counter);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
end;

function TPBKDF2KeyDerivationOLD.GenerateKey(s: string): string;
var
  err, I: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  output, sPB: PAnsiChar;
{$ELSE}
  output, sPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBufferA(s, FUni);
  output := GetMemory(FOutputSizeBits DIV 8);

  for I := 0 to FOutputSizeBits DIV 8 - 1 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
    output[I] := #0;
{$ELSE}
    output[I] := Byte(#0);
{$IFEND}
  try
    err := PBKDF2_KeyDerivationFunction_OLD(sPB, FSalt, FCounter,
      FOutputSizeBits, output);
    if err <> 0 then
    begin
      raise ECryptoPack.Create(err);
    end;
    Result := conv.CharToFormat(output, FOutputSizeBits DIV 8);
  finally
    FreeMem(output);
    FreeMem(sPB);
    conv.Free;
  end;

end;

procedure TPBKDF2KeyDerivationOLD.SetOutputSizeBits(const Value: Integer);
begin
  if Value mod 8 <> 0 then
    raise ECryptoPack.Create
      ('The PBKDF2 output length must be a multiple of 8!');
  FOutputSizeBits := Value;
end;

procedure TPBKDF2KeyDerivationOLD.SetStringSalt(const Value: string);
var
  I, min: Integer;
begin
  FStringSalt := Value;
  if 60 > Length(Value) then
    // to know the length of the salt
    min := Length(Value)
  else
    min := 60;

  for I := 0 to min - 1 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
    FSalt[I] := PAnsiChar(AnsiString(Value))[I];
{$ELSE}
    FSalt[I] := Byte(Value.Chars[I]);
{$IFEND}
  for I := min to 59 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
    FSalt[I] := #0;
{$ELSE}
    FSalt[I] := Byte(#0);
{$IFEND}
end;

procedure TPBKDF2KeyDerivationOLD.SetCounter(const Value: Integer);
begin
  FCounter := Value;
end;

/// / *** new PBKDF2 derivation class **** /////

Constructor TPBKDF2KeyDerivation.Create(AOwner: TComponent);
begin
  self.SetOutputSizeBits(128);
  self.FCounter := 10000;
  self.FOutputFormat := hexa;
  self.FHashFunction := hsha2;
  self.SetHashSizeBits(256);
  self.FUni := yesUni;
  inherited Create(AOwner);
end;

Constructor TPBKDF2KeyDerivation.Create;
begin
  self.SetOutputSizeBits(128);
  self.FCounter := 10000;
  self.FOutputFormat := hexa;
  self.FHashFunction := hsha2;
  self.SetHashSizeBits(256);
  self.FUni := yesUni;
end;

Constructor TPBKDF2KeyDerivation.Create(outputSizeBits: Integer; salt: string;
  counter: Integer; outputFormat: TConvertType; uni: TUnicode;
  hashF: THashFunction; hashSB: Integer);
begin
  self.SetOutputSizeBits(outputSizeBits);
  self.FSalt := salt;
  self.FCounter := counter;
  self.FOutputFormat := outputFormat;
  self.FHashFunction := hashF;
  self.SetHashSizeBits(hashSB);
  self.FUni := uni;
end;

function TPBKDF2KeyDerivation.GenerateKey(s: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  output, sPB, saltPB: PAnsiChar;
{$ELSE}
  output, sPB, saltPB: PByte;
{$IFEND}
  conv: TConvert;
  err, I, saltLen, sLen: Integer;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, sLen);
  saltPB := conv.StringToBuffer(FSalt, FUni, saltLen);
  output := GetMemory(FOutputSizeBits DIV 8);
  for I := 0 to FOutputSizeBits DIV 8 - 1 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
    output[I] := #0;
{$ELSE}
    output[I] := Byte(#0);
{$IFEND}
  try
    err := PBKDF2_KeyDerivationFunction(sPB, sLen, saltPB, saltLen, FCounter,
      Integer(FHashFunction), FHashSizeBits, FOutputSizeBits, output);
    if err <> 0 then
    begin
      raise ECryptoPack.Create(err);
    end;
    Result := conv.CharToFormat(output, FOutputSizeBits DIV 8);
  finally
    FreeMem(output);
    FreeMem(sPB);
    FreeMem(saltPB);
    conv.Free;
  end;

end;

procedure TPBKDF2KeyDerivation.SetOutputSizeBits(const Value: Integer);
begin
  if Value mod 8 <> 0 then
    raise ECryptoPack.Create
      ('The PBKDF2 output length must be a multiple of 8!');
  FOutputSizeBits := Value;
end;

procedure TPBKDF2KeyDerivation.SetHashSizeBits(const Value: Integer);
begin
  if FHashFunction = hsha2 then
  begin
    if (Value <> 256) and (Value <> 512) then
      raise ECryptoPack.Create
        ('With SHA2, the PBKDF2 hashSizeBits parameters should be 256 or 512');
  end
  else
  begin
    if Value mod 8 <> 0 then
      raise ECryptoPack.Create
        ('With SHA3, the PBKDF2 hashSizeBits parameters should be a multiple of 8!');
  end;
  FHashSizeBits := Value;
end;

Constructor TBlake2BHash.Create(AOwner: TComponent);
begin
  self.SetHashSizeBytes(16);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  inherited Create(AOwner);
end;

Constructor TBlake2BHash.Create;
begin
  self.SetHashSizeBytes(16);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
end;

Constructor TBlake2BHash.Create(hashSizeBytes: Integer; key: string;
  outputFormat: TConvertType; uni: TUnicode);
begin
  self.SetHashSizeBytes(hashSizeBytes);
  self.SetKey(key);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
end;

function TBlake2BHash.Hash(s: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, sPB, key: PAnsiChar;
{$ELSE}
  bufferHash, sPB, key: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  key := conv.StringToBufferA(FKey, noUni);
  bufferHash := GetMemory(FHashSizeBytes);
  try
    err := Blake2B_BufferHash(sPB, msgLen, key, FKeylen, bufferHash,
      FHashSizeBytes);
    if err <> 0 then
    begin
      raise ECryptoPack.Create(err);
    end;
    Result := conv.CharToFormat(bufferHash, FHashSizeBytes);
  finally
    FreeMem(bufferHash);
    conv.Free;
    FreeMem(sPB);
    FreeMem(key);
  end;
end;

function TBlake2BHash.HashFile(s: string): string;
var
  stream1: TFileStream;
begin

  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    Result := self.HashStream(stream1);
  finally
    stream1.Free;
  end;

end;

function TBlake2BHash.HashStream(s: TStream): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  key, Hash: PAnsiChar;
  mess: array [0 .. 10000] of AnsiChar;
{$ELSE}
  key, Hash: PByte;
  mess: array [0 .. 10000] of Byte;
{$IFEND}
  state: _blake2b_state;
  nbBlocksRead, I, l: Int64;
  err, nb_bytes, blockLen: Integer;
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  key := conv.StringToBufferA(FKey, noUni);
  Hash := GetMemory(FHashSizeBytes);
  s.Position := 0;
  SetProgress(0);
  blockLen := 10000;
  try
    if FKeylen > 0 then
    begin
      err := initKey_blake2b(@state, FHashSizeBytes, key, FKeylen);
      if err <> 0 then
        raise ECryptoPack.Create(err);
    end
    else
    begin
      err := init_blake2b(@state, FHashSizeBytes);
      if err <> 0 then
        raise ECryptoPack.Create(err);
    end;
    l := s.size;
    I := 0;
    nbBlocksRead := 0;
    while (I < l) do
    begin
      FillChar(mess, 0, blockLen);
      nb_bytes := s.Read(mess[0], blockLen * SizeOf(mess[0]));
      err := update_blake2b(@state, @mess[0], nb_bytes);
      nbBlocksRead := nbBlocksRead + nb_bytes;
      SetProgress((nbBlocksRead * 100) div s.size);
      if err <> 0 then
        raise ECryptoPack.Create(err);
      I := I + nb_bytes;
    end;
    err := finalHash_blake2b(@state, Hash, FHashSizeBytes);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Result := conv.CharToFormat(Hash, FHashSizeBytes);
    SetProgress(0);
  finally
    conv.Free;
    FreeMem(Hash);
    FreeMem(key);
  end;

end;

procedure TBlake2BHash.SetHashSizeBytes(const Value: Integer);
begin
  if (Value > 64) then
    raise ECryptoPack.Create
      ('The size of the hash is incorrect, must be maximum 64');
  FHashSizeBytes := Value;
end;

procedure TBlake2BHash.SetProgress(P: Integer);
begin
  if P <> FProgress then
  begin
    FProgress := P;
    if Assigned(FOnChange) then
      FOnChange(self);
  end;
end;

procedure TBlake2BHash.SetKey(const Value: String);
begin
  if Length(Value) > 64 then
    raise ECryptoPack.Create('The length of the key must be maximum 64');
  FKeylen := Length(Value);
  FKey := Value;
end;

Constructor TRIPEMD160Hash.Create(AOwner: TComponent);
begin
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
  inherited Create(AOwner);
end;

Constructor TRIPEMD160Hash.Create;
begin
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  self.FOnChange := nil;
end;

Constructor TRIPEMD160Hash.Create(outputFormat: TConvertType; uni: TUnicode);
begin
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
  self.FOnChange := nil;
end;

function TRIPEMD160Hash.Hash(s: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  bufferHash, sPB: PAnsiChar;
{$ELSE}
  bufferHash, sPB: PByte;
{$IFEND}
  conv: TConvert;
  msgLen: Integer;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  bufferHash := GetMemory(20);
  try
    RIPEMD160_BufferHash(sPB, msgLen, bufferHash);
    Result := conv.CharToFormat(bufferHash, 20);
  finally
    FreeMem(bufferHash);
    conv.Free;
    FreeMem(sPB);
  end;

end;

function TRIPEMD160Hash.HashFile(s: string): string;
var
  stream1: TFileStream;
begin

  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    Result := self.HashStream(stream1);
  finally
    stream1.Free;
  end;

end;

function TRIPEMD160Hash.HashStream(s: TStream): string;
var
  nbytes, I, j, offset: Integer;
{$IF defined(MACOS64) and not defined(IOS)}
  buf: PInteger;
  X: PInteger;
{$ELSE}
  buf: TRIPEMDBuf;
  X: TRIPEMDX;
{$IFEND}
  nbBlocksRead: Int64;
  Length: array [0 .. 1] of Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  data: array [0 .. 1023] of AnsiChar;
  hashcode: array [0 .. 19] of AnsiChar;
{$ELSE}
  data: array [0 .. 1023] of Byte;
  hashcode: array [0 .. 19] of Byte;
{$IFEND}
  conv: TConvert;
begin
  s.Position := 0;
  SetProgress(0);
  FillChar(Length, 2 * SizeOf(Integer), 0);
  FillChar(data, 1024, 0);
  FillChar(hashcode, 20, 0);
{$IF defined(MACOS64) and not defined(IOS)}
  buf := GetMemory(5 * SizeOf(Integer));
  X := GetMemory(16 * SizeOf(Integer));
  FillChar(buf^, 5 * SizeOf(Integer), 0);
  FillChar(X^, 16 * SizeOf(Integer), 0);
{$ELSE}
  FillChar(buf, 5 * SizeOf(Integer), 0);
  FillChar(X, 16 * SizeOf(Integer), 0);
{$IFEND}
  conv := TConvert.Create;
  try
    conv.AType := outputFormat;
    init_ripemd160(buf);
    nbytes := 1;
    nbBlocksRead := 0;
    while nbytes <> 0 do
    begin
      nbytes := s.Read(data[0], 1024 * SizeOf(data[0]));
      nbBlocksRead := nbBlocksRead + 1024;
      for I := 0 to (nbytes shr 6) - 1 do
      begin
        for j := 0 to 15 do
          X[j] := (Byte(data[64 * I + 4 * j + 3]) shl 24)
            xor (Byte(data[64 * I + 4 * j + 2]) shl 16)
            xor (Byte(data[64 * I + 4 * j + 1]) shl 8)
            xor (Byte(data[64 * I + 4 * j]));
        compress_ripemd160(buf, X);
      end;
      if Length[0] + nbytes < Length[0] then
        Length[1] := Length[1] + 1;
      Length[0] := Length[0] + nbytes;
      SetProgress((nbBlocksRead * 100) div s.size);
    end;

    offset := Length[0] and $3C0;
    finish_ripemd160(buf, @data[offset], Length[0], Length[1]);

    I := 0;
    while I < 20 do
    begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      hashcode[I] := AnsiChar(buf[I shr 2]);
      hashcode[I + 1] := AnsiChar(buf[I shr 2] shr 8);
      hashcode[I + 2] := AnsiChar(buf[I shr 2] shr 16);
      hashcode[I + 3] := AnsiChar(buf[I shr 2] shr 24);
{$ELSE}
      hashcode[I] := buf[I shr 2];
      hashcode[I + 1] := buf[I shr 2] shr 8;
      hashcode[I + 2] := buf[I shr 2] shr 16;
      hashcode[I + 3] := buf[I shr 2] shr 24;
{$IFEND}
      I := I + 4;
    end;
    SetProgress(0);
    Result := conv.CharToFormat(@hashcode[0], 20);
  finally
    conv.Free;
{$IF defined(MACOS64) and not defined(IOS)}
    FreeMem(buf);
    FreeMem(X);
{$IFEND}
  end;

end;

procedure TRIPEMD160Hash.SetProgress(P: Integer);
begin
  if P <> FProgress then
  begin
    FProgress := P;
    if Assigned(FOnChange) then
      FOnChange(self);
  end;
end;

Constructor TArgon2KeyDerivation.Create(AOwner: TComponent);
begin
  self.SetOutputSizeBytes(16);
  self.SetCounter(10);
  self.SetMemory(16);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
  inherited Create(AOwner);
end;

Constructor TArgon2KeyDerivation.Create;
begin
  self.SetOutputSizeBytes(16);
  self.SetCounter(10);
  self.SetMemory(16);
  self.FOutputFormat := hexa;
  self.FUni := yesUni;
end;

Constructor TArgon2KeyDerivation.Create(outputSizeBytes: Integer; salt: string;
  counter: Integer; outputFormat: TConvertType; memory: Integer; uni: TUnicode);
begin
  self.SetOutputSizeBytes(outputSizeBytes);
  self.SetStringSalt(salt);
  self.SetCounter(counter);
  self.SetMemory(memory);
  self.FOutputFormat := outputFormat;
  self.FUni := uni;
end;

destructor TArgon2KeyDerivation.Destroy;
begin
  if Assigned(FSalt) then
  begin
    FreeMem(FSalt);
    FSalt := nil;
  end;
  inherited Destroy;
end;

function TArgon2KeyDerivation.GenerateKey(s: string): string;
var
  err, msgLen: Integer;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  output, sPB: PAnsiChar;
  I: Integer;
{$ELSE}
  output, sPB: PByte;
{$IFEND}
  conv: TConvert;
begin
  conv := TConvert.Create(outputFormat);
  sPB := conv.StringToBuffer(s, FUni, msgLen);
  output := GetMemory(FOutputSizeBytes + 1);
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  for I := 0 to FOutputSizeBytes do
    output[I] := #0;
{$ELSE}
  FillChar(output^, FOutputSizeBytes, Byte(#0));
{$IFEND}
  try
    err := Argon2_KeyDerivation(FCounter, FMemory, sPB, msgLen, FSalt, output,
      FOutputSizeBytes);
    if err <> 0 then
    begin
      raise ECryptoPack.Create(err);
    end;
    Result := conv.CharToFormat(output, FOutputSizeBytes);
  finally
    FreeMem(output);
    conv.Free;
    FreeMem(sPB);
  end;

end;

procedure TArgon2KeyDerivation.SetOutputSizeBytes(const Value: Integer);
begin
  if Value < 4 then
    raise ECryptoPack.Create('The Argon2 output length must be minimum 4!');
  FOutputSizeBytes := Value;
end;

procedure TArgon2KeyDerivation.SetStringSalt(const Value: string);
var
  I, min: Integer;
  conv: TConvert;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  sb: PAnsiChar;
{$ELSE}
  sb: PByte;
{$IFEND}
begin
  if Assigned(FSalt) then
    FreeMem(FSalt);
  FStringSalt := Value;
  conv := TConvert.Create;
  try
    sb := conv.StringToBuffer(FStringSalt, noUni, min);
  finally
    conv.Free;
  end;

  if 16 > Length(Value) then
    // to know the length of the salt
    min := Length(Value)
  else
    min := 16;

  try
    FSalt := GetMemory(17);
    for I := 0 to min - 1 do
    begin
      FSalt[I] := sb[I];
    end;

    for I := min to 16 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      FSalt[I] := #0;
{$ELSE}
      FSalt[I] := Byte(#0);
{$IFEND}
  finally
    FreeMem(sb);
  end;
end;

procedure TArgon2KeyDerivation.SetCounter(const Value: Integer);
begin
  if Value < 1 then
    raise ECryptoPack.Create('The Argon2 counter must be minimum 1!');
  FCounter := Value;
end;

procedure TArgon2KeyDerivation.SetMemory(const Value: Integer);
begin
  if Value < 8 then
    raise ECryptoPack.Create('The Argon2 memory must be minimum 8!');
  FMemory := Value;
end;

procedure THKDFKeyDerivation.SetHashSizeBits(const Value: Integer);
begin
  if FHashFunction = hsha2 then
  begin
    if (Value <> 256) and (Value <> 512) and (Value <> 384) then
      raise ECryptoPack.Create
        ('With SHA2, the HKDF hashSizeBits parameters should be 256, 384 or 512');
  end
  else
  begin
    if Value mod 8 <> 0 then
      raise ECryptoPack.Create
        ('With SHA3, the HKDF hashSizeBits parameters should be a multiple of 8!');
  end;
  FHashSizeBits := Value;
end;

Constructor THKDFKeyDerivation.Create(AOwner: TComponent);
begin
  self.FOutputFormat := hexa;
  self.FHashFunction := hsha2;
  self.SetHashSizeBits(256);
  self.FUni := yesUni;
  inherited Create(AOwner);
end;

Constructor THKDFKeyDerivation.Create;
begin
  self.FOutputFormat := hexa;
  self.FHashFunction := hsha2;
  self.SetHashSizeBits(256);
  self.FUni := yesUni;
end;

Constructor THKDFKeyDerivation.Create(outputFormat: TConvertType; uni: TUnicode;
  hashF: THashFunction; hashSB: Integer);
begin
  self.FOutputFormat := outputFormat;
  self.FHashFunction := hashF;
  self.SetHashSizeBits(hashSB);
  self.FUni := uni;
end;

function THKDFKeyDerivation.Expand(s, info: string; len: Integer): string;
var
  N, I: Integer;
  T: array of string;
  SHA2: TSHA2Hash;
  SHA3: TSHA3Hash;
  conv: TConvert;
begin
  N := ceil(len / (self.hashSizeBits div 8));

  SetLength(T, N + 1);
  T[0] := '';

  if self.hashFunction = hsha2 then
  begin
    SHA2 := TSHA2Hash.Create(self.hashSizeBits, raw, self.Unicode);
    try
      for I := 1 to N do
      begin
        if I > 1 then
          SHA2.Unicode := noUni;
        T[I] := SHA2.HMAC(T[I - 1] + info + Char(I), s);
      end;

    finally
      SHA2.Free;
    end;

  end
  else
  begin
    SHA3 := TSHA3Hash.Create(self.hashSizeBits, raw, self.Unicode);
    try

      for I := 1 to N do
      begin
        T[I] := SHA3.HMAC(T[I - 1] + info + Char(I), s);
      end;

    finally
      SHA3.Free;
    end;
  end;

  Result := '';

  for I := 1 to N do
    Result := Result + T[I];

  Result := copy(Result, 0, len);

  conv := TConvert.Create(self.outputFormat);
  try
    Result := conv.StringToFormat(Result);
  finally
    conv.Free;
  end;

end;

function THKDFKeyDerivation.Extract(s, salt: string): string;
var
  SHA2: TSHA2Hash;
  SHA3: TSHA3Hash;
begin
  if self.hashFunction = hsha2 then
  begin
    SHA2 := TSHA2Hash.Create(self.hashSizeBits, self.outputFormat,
      self.Unicode);
    try
      Result := SHA2.HMAC(s, salt);
    finally
      SHA2.Free;
    end;
  end
  else
  begin
    SHA3 := TSHA3Hash.Create(self.hashSizeBits, self.outputFormat,
      self.Unicode);
    try
      Result := SHA3.HMAC(s, salt);
    finally
      SHA3.Free;
    end;
  end;

end;

{$IF (defined(WIN32) or defined(MACOS32)) and (not defined(IOS))}

procedure __llshl;
asm
  jmp System.@_llshl
end;

procedure __llushr;
asm
  jmp System.@_llushr
end;

procedure __llmul;
asm
  jmp System.@_llmul
end;

procedure __llumod;
asm
  jmp System.@_llumod
end;

procedure __lludiv;
asm
  jmp System.@_lludiv
end;

procedure __lldiv;
asm
  jmp System.@_lldiv
end;

procedure __aulldiv; external 'ntdll' name '_aulldiv';

procedure __alldiv; external 'ntdll' name '_alldiv';

procedure __aullrem; external 'ntdll' name '_aullrem';

function _wcscpy(destination: PChar; source: PChar): PChar;
  external msvcrt name 'wcscpy';

{$IFEND}
{$IF defined(WIN32)}
procedure __chkstk_noalloc; external 'ntdll';
{$IFEND}

end.
