{ *************************************************************************** }
{ TMS Cryptography Pack }
{ for Delphi & C++Builder }
{ }
{ Written by Cyberens Technologies & Services }
{ Exclusively commercialized by tmssoftware.com bvba }
{ copyright  2016 - 2020 by Cyberens Technologies & Services }
{ Email : info@tmssoftware.com }
{ Web : http://www.tmssoftware.com }
{ }
{ The source code is given as is. The author is not responsible }
{ for any possible damage done due to the use of this code. }
{ The component can be freely used in any application. The complete }
{ source code remains property of the author and may not be distributed, }
{ published, given or sold in any form as such. No parts of the source }
{ code can be included in any other component or application without }
{ written authorization of the author. }
{ *************************************************************************** }

unit MiscObj;

interface

{$I TMSDEFS.INC}
{$I TMSCRYPTO.INC}

uses
  CryptBase, SysUtils, Classes;

type

  TConvertType = (base64, hexa, base64url, base32, raw);
  TUnicode = (noUni, yesUni);

  TConvert = class(TTMSCryptBase)
  private
    FType: TConvertType;
    function NormalizeFormat(str: string): string;
  public
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create; reintroduce; overload;
    constructor Create(AType: TConvertType); reintroduce; overload;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
    function CharToFormat(charstring: PAnsiChar; charlen: Integer): string;
    function FormatToChar(str: string): PAnsiChar;
    function UnicodeToPAnsiChar(str: string): PAnsiChar;
    function PAnsiCharFromUnicodeLength(str: string): Integer;
    function StringToBuffer(str: string; u: TUnicode; var msgLen: Integer)
      : PAnsiChar;
    function StringToBufferA(str: string; u: TUnicode): PAnsiChar;
    function StringToPAnsiChar(str: string): PAnsiChar;
{$ELSE}
    function CharToFormat(charstring: PByte; charlen: Integer): string;
    function FormatToChar(str: string): PByte;
    function StringToBuffer(str: string; u: TUnicode;
      var msgLen: Integer): PByte;

    function StringToBufferA(str: string; u: TUnicode): PByte;
{$IFEND}
    function UnicodeStringToFormat(str: string): string;
    function FormatToUnicodeString(str: string): string;
    function StringToByteArray(str: string): TArray<Byte>;
    function TestUnicode(str: string): Integer;
    function StringToUnicode(str: string): string;
    function StringToFormat(charstring: string): string;
    function FormatToString(str: string): string;
    function OutputFormatLength(charlen: Integer): Integer;
    function CharLength(charstring: string): Integer;
    function Base64ToHexa(base64String: string): string;
    function HexaToBase64(hexaString: string): string;
    function Base64ToBase64url(inString: string): string;
    function Base64urlToBase64(inString: string): string;
    function Base64urlToHexa(inString: string): string;
    function HexaToBase64url(inString: string): string;
    function Base32ToHexa(base32String: string): string;
    function HexaToBase32(hexaString: string): string;
    function Base32ToBase64url(inString: string): string;
    function Base64urlToBase32(inString: string): string;
    function Base32ToBase64(inString: string): string;
    function Base64ToBase32(inString: string): string;
    function KeyRSAOpenSSLToKeyTRSAEncSign(strKey: string): string;
    function KeyTRSAEncSignToKeyRSAOpenSSL(strKey: string): string;
    function Base58Encode(const value: uint64): string;
    function Base58Decode(const encoded: string): uint64;
    function TBytesToString(const t: TBytes): string;
    function StringToTBytes(const str: string): TBytes;
    function RandomString(len: Integer): string;
  published
    property AType: TConvertType read FType write FType default hexa;
  end;

type
  ECryptoPack = class(Exception)
  private
    FCode: Integer;
  public
    constructor Create(Code: Integer); overload;
    constructor Create(errorMessag: string); overload;
    property Code: Integer read FCode;
  end;

function PosExZ(const SubString: string; const s: string;
  const StartIndex: Integer): Integer;

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
function CharToHexa(charstring: PAnsiChar; lenCharString: Integer;
  hexaString: PAnsiChar): Integer; cdecl;
function HexaToChar(hexaString: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl;
function Base64ToChar(base64String: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl;
function CharToBase64(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl;
function Base32ToChar(base32String: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl;
function CharToBase32(charstring: PAnsiChar; lenCharString: Integer;
  base32String: PAnsiChar): Integer; cdecl;
function Base64urlToChar(base64String: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl;
function CharToBase64url(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl;
procedure ErrorMessage(error: Integer; message: PAnsiChar); cdecl;
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
function RandomInt(): Integer; cdecl;
function RandomUInt(): Integer; cdecl;
{$ELSE}
function CharToHexa(charstring: PByte; lenCharString: Integer;
  hexaString: PByte): Integer;
function HexaToChar(hexaString: PByte; charstring: PByte): Integer;
function Base64ToChar(base64String: PByte; charstring: PByte): Integer;
function CharToBase64(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer;
function Base32ToChar(base32String: PByte; charstring: PByte): Integer;
function CharToBase32(charstring: PByte; lenCharString: Integer;
  base32String: PByte): Integer;
function Base64urlToChar(base64String: PByte; charstring: PByte): Integer;
function CharToBase64url(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer;
procedure ErrorMessage(error: Integer; message: array of Byte);
function ShortStringToString(value: PByte; I: Integer): String;
procedure StringToPByte(str: string; var output: PByte);
function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
function RandomInt(): Integer;
function RandomUInt(): Integer;
{$IFEND}

implementation

uses
  Math, DateUtils,
{$IFDEF MSWINDOWS}
  Winapi.Windows, AnsiStrings, System.Win.Crtl,
{$ENDIF}
{$IFDEF MACOS}
  System.Mac.CFUtils,
{$ENDIF}
{$IFDEF POSIX}
  Posix.Dlfcn, Posix.Fcntl, Posix.SysStat, Posix.SysTime, Posix.SysTypes,
  Posix.Unistd, Posix.Stdio, Posix.Stdlib,
{$ENDIF POSIX}
{$IFDEF PC_MAPPED_EXCEPTIONS}
  System.Internal.Unwinder,
{$ENDIF PC_MAPPED_EXCEPTIONS}
  System.SysConst;

const
  BASE58_ALPHABET =
    '123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ';
  msvcrt = 'msvcrt.dll';
  ADVAPI32 = 'advapi32.dll';
{$IFNDEF WINXP}
  BCRYPT = 'bcrypt.dll';
{$ENDIF}
{$IF not defined(IDEVERSION1021)}
  RANDOMDLL = 'RandomDLL.dll'; // Copy the DLL from Win64 directory
  // into C:\Windows\SysWOW64 if you are running 32 bit Windows
  // or into C:\Windows\System32 if you are running 64 bit Windows
  // If you are using C++ demo in Win64 target, add this line to MiscObj.hpp
  // #pragma comment(lib, "random-dll.o")
  // and put the directory where is this file in Project->Options->Search Path
{$IFEND}
  libc = '/usr/lib/libc.dylib';
  PROV_RSA_FULL = 1;
  CRYPT_VERIFYCONTEXT = $F0000000;
{$IF (defined(WIN32) or defined(MACOS)) and (not defined(IOS))}
  PU = '_';
{$ELSE}
  PU = '';
{$IFEND}
{$IF defined(WIN32)}
{$IFDEF WINXP}
{$LINK './libWIN32/elt_rep_XP.obj'}
{$ELSE}
{$LINK './libWIN32/elt_rep.obj'}
{$ENDIF}
{$LINK './libWIN32/error.obj'}
{$ELSEIF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
{$LINK './libWIN64/elt_rep.o'}
{$LINK './libWIN64/error.o'}
{$ELSE}
{$LINK './libWIN64/elt_rep.obj'}
{$LINK './libWIN64/error.obj'}
{$IFEND}
{$ELSEIF defined(MACOS32) and not defined(IOS)}
{$LINK './libOSX32/elt_rep.o'}
{$LINK './libOSX32/error.o'}
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS32)) and (not defined(IOS))}
function CharToHexa(charstring: PAnsiChar; lenCharString: Integer;
  hexaString: PAnsiChar): Integer; cdecl; external name PU + 'CharToHexa';
function HexaToChar(hexaString: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external name PU + 'HexaToChar';
function Base64ToChar(base64String: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external name PU + 'Base64ToChar';
function CharToBase64(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl; external name PU + 'CharToBase64';
function Base64urlToChar(base64String: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl; external name PU + 'Base64urlToChar';

function CharToBase64url(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl;
  external name PU + 'CharToBase64url';
function Base32ToChar(base32String: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external name PU + 'Base32ToChar';
function CharToBase32(charstring: PAnsiChar; lenCharString: Integer;
  base32String: PAnsiChar): Integer; cdecl; external name PU + 'CharToBase32';
procedure ErrorMessage(error: Integer; message: PAnsiChar); cdecl;
  external name PU + 'ErrorMessage';
{$IF defined(MACOS) or defined(WIN32)}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomUBuffer';
function RandomInt(): Integer; cdecl; external name PU + 'RandomInt';
function RandomUInt(): Integer; cdecl; external name PU + 'RandomUInt';
{$IFEND}
{$IFDEF WIN32}
{$IFDEF WINXP}
function CryptAcquireContextW(phProv: pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
{$ELSE}
function BCryptOpenAlgorithmProvider(out phAlgorithm: THandle;
  pszAlgId: LPCWSTR; pszProvider: LPCWSTR; dwProvType: DWORD; dwFlags: DWORD)
  : LongInt; stdcall; external BCRYPT name 'BCryptOpenAlgorithmProvider';
function BCryptCloseAlgorithmProvider(hAlgorithm: THandle; dwFlags: ULONG)
  : LongInt; stdcall; external BCRYPT name 'BCryptCloseAlgorithmProvider';
function BCryptGenRandom(hAlgorithm: THandle; out pbBuffer: PUCHAR;
  cbBuffer, dwFlags: ULONG): LongInt; stdcall;
  external BCRYPT name 'BCryptGenRandom';
{$ENDIF}
{$ELSEIF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
function CryptAcquireContextW(phProv: pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomBuffer';
function RandomInt(): Integer; external name PU + 'RandomInt';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomUBuffer';
function RandomUInt(): Integer; external name PU + 'RandomUInt';
{$ELSE}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external RANDOMDLL name 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external RANDOMDLL name 'RandomUBuffer';
function RandomInt(): Integer; cdecl; external RANDOMDLL name 'RandomInt';
function RandomUInt(): Integer; cdecl; external RANDOMDLL name 'RandomUInt';
{$IFEND}
{$IFEND}
{$IFDEF MACOS}
function _open(PathName: MarshaledAString; Flags: Integer): Integer; cdecl;
  varargs; external libc name PU + 'open';
function _read(Handle: Integer; Buffer: pointer; Count: size_t): ssize_t; cdecl;
  external libc name PU + 'read';
function _close(filedes: Integer): Integer; cdecl;
  external libc name PU + 'close';
function _memset(var b: pointer; c: Integer; len: size_t): pointer; cdecl;
  external libc name PU + 'memset';
function _malloc(size: size_t): pointer; cdecl;
  external libc name PU + 'malloc';
procedure _memcpy(dest, source: pointer; Count: Integer); cdecl;
  external libc name PU + 'memcpy';
procedure _free(p: pointer); cdecl; external libc name PU + 'free';
function _realloc(p: pointer; NewSize: size_t): pointer; cdecl;
  external libc name PU + 'realloc';
function _strlen(S1: MarshaledAString): size_t; cdecl;
  external libc name PU + 'strlen';
function _strcat(S1: MarshaledAString; s2: MarshaledAString): MarshaledAString;
  cdecl; external libc name PU + 'strcat';
{$ENDIF}
{$ELSEIF defined(ANDROID) or defined(IOS)}
function CharToHexa(charstring: PByte; lenCharString: Integer;
  hexaString: PByte): Integer; external 'libTMSCPLib.a';
function HexaToChar(hexaString: PByte; charstring: PByte): Integer;
  external 'libTMSCPLib.a';
function Base64ToChar(base64String: PByte; charstring: PByte): Integer;
  external 'libTMSCPLib.a';
function CharToBase64(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer; external 'libTMSCPLib.a';
function Base32ToChar(base32String: PByte; charstring: PByte): Integer;
  external 'libTMSCPLib.a';
function CharToBase32(charstring: PByte; lenCharString: Integer;
  base32String: PByte): Integer; external 'libTMSCPLib.a';
function Base64urlToChar(base64String: PByte; charstring: PByte): Integer;
  external 'libTMSCPLib.a';
function CharToBase64url(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer; external 'libTMSCPLib.a';
procedure ErrorMessage(error: Integer; message: array of Byte);
  external 'libTMSCPLib.a';
function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomInt(): Integer; external 'libTMSCPLib.a';
function RandomUInt(): Integer; external 'libTMSCPLib.a';
{$ELSEIF defined(LINUX)}
function CharToHexa(charstring: PByte; lenCharString: Integer;
  hexaString: PByte): Integer; external 'libTMSCPlib.a';
function HexaToChar(hexaString: PByte; charstring: PByte): Integer;
  external 'libTMSCPlib.a';
function Base64ToChar(base64String: PByte; charstring: PByte): Integer;
  external 'libTMSCPlib.a';
function CharToBase64(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer; external 'libTMSCPlib.a';
function Base32ToChar(base32String: PByte; charstring: PByte): Integer;
  external 'libTMSCPlib.a';
function CharToBase32(charstring: PByte; lenCharString: Integer;
  base32String: PByte): Integer; external 'libTMSCPlib.a';
function Base64urlToChar(base64String: PByte; charstring: PByte): Integer;
  external 'libTMSCPlib.a';
function CharToBase64url(charstring: PByte; lenCharString: Integer;
  base64String: PByte): Integer; external 'libTMSCPlib.a';
procedure ErrorMessage(error: Integer; message: array of Byte);
  external 'libTMSCPlib.a';
function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomInt(): Integer; external 'libTMSCPlib.a';
function RandomUInt(): Integer; external 'libTMSCPlib.a';
{$ELSE}
function CharToHexa(charstring: PAnsiChar; lenCharString: Integer;
  hexaString: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'CharToHexa';
function HexaToChar(hexaString: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/elt_rep.o' name 'HexaToChar';
function Base64ToChar(base64String: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/elt_rep.o' name 'Base64ToChar';
function CharToBase64(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'CharToBase64';
function Base64urlToChar(base64String: PAnsiChar; charstring: PAnsiChar)
  : Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'Base64urlToChar';

function CharToBase64url(charstring: PAnsiChar; lenCharString: Integer;
  base64String: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'CharToBase64url';
function Base32ToChar(base32String: PAnsiChar; charstring: PAnsiChar): Integer;
  cdecl; external object'./libOSX64/elt_rep.o' name 'Base32ToChar';
function CharToBase32(charstring: PAnsiChar; lenCharString: Integer;
  base32String: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'CharToBase32';
procedure ErrorMessage(error: Integer; message: PAnsiChar); cdecl;
  external object'./libOSX64/error.o' name 'ErrorMessage';
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'RandomUBuffer';
function RandomInt(): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'RandomInt';
function RandomUInt(): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'RandomUInt';
{$IFEND}

constructor TConvert.Create(AOwner: TComponent);
begin
  FType := hexa;
  inherited Create(AOwner);
end;

constructor TConvert.Create;
begin
  FType := hexa;
end;

constructor TConvert.Create(AType: TConvertType);
begin
  FType := AType;
end;

constructor ECryptoPack.Create(errorMessag: string);
begin
  inherited Create(errorMessag);
end;

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

constructor ECryptoPack.Create(Code: Integer);
var
  errorMsg: array [0 .. 2000] of AnsiChar;
begin
  ErrorMessage(Code, errorMsg);
  inherited Create(string(errorMsg));
  FCode := Code;
end;

function TConvert.CharToFormat(charstring: PAnsiChar; charlen: Integer): string;
var
  s: PAnsiChar;
  err, L, I: Integer;
begin
  if FType <> raw then
  begin
    L := OutputFormatLength(charlen) * SizeOf(AnsiChar);
    s := GetMemory(L);
    try
      if FType = hexa then
        err := CharToHexa(charstring, charlen, s)
      else if FType = base64 then
        err := CharToBase64(charstring, charlen, s)
      else if FType = base64url then
        err := CharToBase64url(charstring, charlen, s)
      else if FType = base32 then
        err := CharToBase32(charstring, charlen, s)
      else
        raise ECryptoPack.CreateFmt('Bad Type (%d)', [Ord(FType)]);

      if err <> 0 then
        raise ECryptoPack.Create(err);

      SetString(Result, s, L - 1);
    finally
      FreeMem(s);
    end;
  end
  else { FType = raw }
  begin
    Result := '';
    for I := 0 to charlen - 1 do
      Result := Result + Char(Ord(charstring[I]));
  end
end;

function TConvert.FormatToChar(str: string): PAnsiChar;
var
  s, t: PAnsiChar;
  err, I, len: Integer;
begin
  if FType <> raw then
  begin
    str := NormalizeFormat(str);
    t := GetMemory((Length(str) + 1) * SizeOf(AnsiChar));
    for I := 0 to Length(str) do
      t[I] := #0;
    try
      for I := 1 to Length(str) do
        t[I - 1] := AnsiChar(Ord(str[I]));

      len := CharLength(str);
      s := GetMemory((len + 1) * SizeOf(AnsiChar));
      for I := 0 to len do
        s[I] := #0;

      if FType = hexa then
        err := HexaToChar(t, s)
      else if FType = base64 then
        err := Base64ToChar(t, s)
      else if FType = base64url then
        err := Base64urlToChar(t, s)
      else if FType = base32 then
        err := Base32ToChar(t, s)
      else
        raise ECryptoPack.CreateFmt('Bad Type (%d)', [Ord(FType)]);
      if err <> 0 then
      begin
        FreeMem(s);
        raise ECryptoPack.Create(err);
      end;
    finally
      FreeMem(t);
    end;
  end
  else
  begin
    s := GetMemory((Length(str) + 1) * SizeOf(AnsiChar));
    for I := 0 to Length(str) do
      s[I] := #0;
    for I := 1 to Length(str) do
      s[I - 1] := AnsiChar(str[I - 1]);
  end;
  Result := s;
end;

function TConvert.StringToPAnsiChar(str: string): PAnsiChar;
var
  t: TArray<Byte>;
begin
  t := StringToByteArray(str);
  Result := GetMemory((Length(t) + 1) * SizeOf(AnsiChar));
  Move(t[Low(t)], Result^, Length(t));
  Result[High(t) + 1] := #0;
end;

function TConvert.UnicodeToPAnsiChar(str: string): PAnsiChar;
var
  t: TArray<Byte>;
begin
  t := TEncoding.UTF8.GetBytes(str);
  Result := GetMemory((Length(t) + 1) * SizeOf(AnsiChar));
  Move(t[Low(t)], Result^, Length(t));
  Result[High(t) + 1] := #0;
end;

function TConvert.PAnsiCharFromUnicodeLength(str: string): Integer;
begin
  Result := TEncoding.UTF8.GetByteCount(str);
end;

{$ELSE}

constructor ECryptoPack.Create(Code: Integer);
var
  errorMsg: array [0 .. 2000] of Byte;
begin
  ErrorMessage(Code, errorMsg);
  inherited Create(ShortStringToString(@errorMsg, 2000));
  FCode := Code;
end;

function TConvert.CharToFormat(charstring: PByte; charlen: Integer): string;
var
  s: PByte;
  err, L, I: Integer;
begin
  if FType <> raw then
  begin
    L := OutputFormatLength(charlen) * SizeOf(Byte);
    s := GetMemory(L);
    try
      if FType = hexa then
        err := CharToHexa(charstring, charlen, s)
      else if FType = base64 then
        err := CharToBase64(charstring, charlen, s)
      else if FType = base64url then
        err := CharToBase64url(charstring, charlen, s)
      else if FType = base32 then
        err := CharToBase32(charstring, charlen, s)
      else
        raise ECryptoPack.CreateFmt('Bad Type (%d)', [Ord(FType)]);

      if err <> 0 then
        raise ECryptoPack.Create(err);

      Result := '';
      for I := 0 to L - 2 do
        Result := Result + Char(s[I]);

    finally
      FreeMem(s);
    end;
  end
  else { FType = raw }
  begin
    Result := '';
    for I := 0 to charlen - 1 do
      Result := Result + Char(charstring[I]);
  end;
end;

function TConvert.FormatToChar(str: string): PByte;
var
  s, t: PByte;
  I, len: Integer;
  err: Integer;
begin
  str := NormalizeFormat(str);
  if Length(str) > 0 then
  begin
    t := GetMemory((Length(str) + 1) * SizeOf(Byte));
    for I := 0 to Length(str) do
      t[I] := Byte(#0);
    for I := 0 to Length(str) - 1 do
{$IF defined(LINUX)}
      t[I] := Byte(str[I + 1]);
{$ELSE}
      t[I] := Byte(str.Chars[I]);
{$IFEND}
  end
  else
  begin
    t := GetMemory(SizeOf(Byte));
    t[0] := Byte(#0);
  end;

  len := CharLength(str);
  s := GetMemory((len + 1) * SizeOf(Byte));
  for I := 0 to len do
    s[I] := Byte(#0);
  try
    try
      if FType = hexa then
        err := HexaToChar(t, s)
      else if FType = base64 then
        err := Base64ToChar(t, s)
      else if FType = base64url then
        err := Base64urlToChar(t, s)
      else if FType = base32 then
        err := Base32ToChar(t, s)
      else
      begin
        err := 0;
        for I := 0 to Length(str) - 1 do
          s[I] := t[I];
      end;
      if err <> 0 then
        raise ECryptoPack.Create(err);
    except
      FreeMem(s);
    end;
  finally
    Result := s;
    FreeMem(t);
  end;
end;

procedure StringToPByte(str: string; var output: PByte);
var
  I, r, f: Integer;
begin
  for I := 0 to Length(str) - 1 do
  begin
{$IF defined(LINUX)}
    r := Ord(str[I + 1]);
{$ELSE}
    r := Ord(str.Chars[I]);
{$IFEND}
    f := Ord(output[I]);
    if r < 255 then
    begin
      if r <> f then
        output[I] := r;
    end;
  end;
end;

function ShortStringToString(value: PByte; I: Integer): String;
var
  b: TBytes;
  L: Integer;
  J: Integer;
begin
  Result := '';
  L := I;
  SetLength(b, L);
  if L > 0 then
  begin
    for J := Low(b) to High(b) do
      b[J] := Byte(#0);
  end;
  Move(value[0], b[0], L);
  Result := TEncoding.ANSI.GetString(b);
end;

{$IFEND}

function TConvert.NormalizeFormat(str: string): string;
var
  len: Integer;
begin
  len := Length(str);
  if (FType = base64) or (FType = base64url) then
  begin
    while len mod 4 <> 0 do
    begin
      str := str + '=';
      len := Length(str);
    end;
  end
  else if FType = base32 then
  begin
    while len mod 8 <> 0 do
    begin
      str := str + '=';
      len := Length(str);
    end;
  end;
  Result := str;
end;

function TConvert.TestUnicode(str: string): Integer;
var
  I: Integer;
begin
  Result := 0;
{$IF defined(ANDROID) or defined(IOS)}
  for I := 0 to Length(str) - 1 do
{$ELSE}
  for I := 1 to Length(str) do
{$IFEND}
  begin
    if Ord(str[I]) > 255 then
    begin
      Result := 1;
      Exit;
    end;
  end;
end;

function TConvert.StringToUnicode(str: string): string;
var
  t: TArray<Byte>;
begin
  t := StringToByteArray(str);
  Result := TEncoding.UTF8.GetString(t);
end;

function TConvert.OutputFormatLength(charlen: Integer): Integer;
begin
  if FType = hexa then
    Result := charlen * 2 + 1
  else if FType = base32 then
  begin
    if charlen mod 5 = 0 then
      Result := (charlen DIV 5) * 8 + 1
    else
      Result := (charlen DIV 5 + 1) * 8 + 1;
  end
  else if (FType = base64) or (FType = base64url) then
  begin
    if charlen mod 3 = 0 then
      Result := (charlen DIV 3) * 4 + 1
    else
      Result := (charlen DIV 3 + 1) * 4 + 1;

  end
  else
  begin
    Result := charlen + 1;
  end;
end;

function TConvert.CharLength(charstring: string): Integer;
var
  len: Integer;
begin
  charstring := NormalizeFormat(charstring);
  Result := 1;
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
  len := Length(charstring) + 1;
  if len = 1 then
    Result := 0;
{$ELSE}
  len := Length(charstring);
  if len = 0 then
    Result := 0;
{$IFEND}
  if Result <> 0 then
  begin
    if FType = hexa then
    begin
      Result := len DIV 2
    end
    else if FType = base32 then
    begin
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
      if charstring[len - 1] <> '=' then
        Result := (len DIV 8) * 5
      else if charstring[len - 2] <> '=' then
        Result := (len DIV 8) * 5 - 1
      else if charstring[len - 4] <> '=' then
        Result := (len DIV 8) * 5 - 2
      else if charstring[len - 5] <> '=' then
        Result := (len DIV 8) * 5 - 3
      else
        Result := (len DIV 8) * 5 - 4;
{$ELSE}
      if charstring.Chars[len - 1] <> '=' then
        Result := (len DIV 8) * 5
      else if charstring.Chars[len - 2] <> '=' then
        Result := (len DIV 8) * 5 - 1
      else if charstring.Chars[len - 4] <> '=' then
        Result := (len DIV 8) * 5 - 2
      else if charstring.Chars[len - 5] <> '=' then
        Result := (len DIV 8) * 5 - 3
      else
        Result := (len DIV 8) * 5 - 4;
{$IFEND}
    end
    else if (FType = base64) or (FType = base64url) then
    begin
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
      if charstring[len - 1] <> '=' then
        Result := (len DIV 4) * 3
      else if charstring[len - 2] <> '=' then
        Result := (len DIV 4) * 3 - 1
      else
        Result := (len DIV 4) * 3 - 2;
{$ELSE}
      if charstring.Chars[len - 1] <> '=' then
        Result := (len DIV 4) * 3
      else if charstring.Chars[len - 2] <> '=' then
        Result := (len DIV 4) * 3 - 1
      else
        Result := (len DIV 4) * 3 - 2;
{$IFEND}
    end
    else
    begin
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
      Result := len - 1;
{$ELSE}
      Result := len;
{$IFEND}
    end;
  end;
end;

function TConvert.Base64ToHexa(base64String: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;
  AType := base64;
  charlen := Length(base64String);
  while charlen mod 4 <> 0 do
  begin
    base64String := base64String + '=';
    charlen := Length(base64String);
  end;
  charlen := CharLength(base64String);
  charstring := FormatToChar(base64String);
  try
    AType := hexa;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base64urlToHexa(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;
  AType := base64url;
  charlen := Length(inString);
  while charlen mod 4 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := hexa;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base64urlToBase64(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base64url;
  charlen := Length(inString);
  while charlen mod 4 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base64;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.HexaToBase64(hexaString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := hexa;
  charlen := CharLength(hexaString);
  charstring := FormatToChar(hexaString);
  try
    AType := base64;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.HexaToBase64url(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := hexa;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base64url;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base64ToBase64url(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base64;
  charlen := Length(inString);
  while charlen mod 4 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base64url;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.KeyRSAOpenSSLToKeyTRSAEncSign(strKey: string): string;
var
  str, str2: string;
  I: Integer;
begin
  if AType = hexa then
  begin
{$IF defined(ANDROID) or defined(IOS)}
    I := 1;
{$ELSE}
    I := 0;
{$IFEND}
    while I < Length(strKey) do
    begin
{$IF defined(ANDROID) or defined(IOS)}
      str := str + strKey.Chars[Length(strKey) - I - 1];
      str := str + strKey.Chars[Length(strKey) - I];
{$ELSE}
      str := str + strKey[Length(strKey) - I - 1];
      str := str + strKey[Length(strKey) - I];
{$IFEND}
      I := I + 2;
    end;
    Result := str;
  end
  else if AType = base64 then
  begin
    str := Base64ToHexa(strKey);
{$IF defined(ANDROID) or defined(IOS)}
    I := 1;
{$ELSE}
    I := 0;
{$IFEND}
    while I < Length(str) do
    begin
{$IF defined(ANDROID) or defined(IOS)}
      str2 := str2 + str.Chars[Length(str) - I - 1];
      str2 := str2 + str.Chars[Length(str) - I];
{$ELSE}
      str2 := str2 + str[Length(str) - I - 1];
      str2 := str2 + str[Length(str) - I];
{$IFEND}
      I := I + 2;
    end;
    Result := HexaToBase64(str2);
  end
  else if AType = base64url then
  begin
    str := Base64urlToHexa(strKey);
{$IF defined(ANDROID) or defined(IOS)}
    I := 1;
{$ELSE}
    I := 0;
{$IFEND}
    while I < Length(str) do
    begin
{$IF defined(ANDROID) or defined(IOS)}
      str2 := str2 + str.Chars[Length(str) - I - 1];
      str2 := str2 + str.Chars[Length(str) - I];
{$ELSE}
      str2 := str2 + str[Length(str) - I - 1];
      str2 := str2 + str[Length(str) - I];
{$IFEND}
      I := I + 2;
    end;
    Result := HexaToBase64url(str2);
  end
  else if AType = base32 then
  begin
    str := Base32ToHexa(strKey);
{$IF defined(ANDROID) or defined(IOS)}
    I := 1;
{$ELSE}
    I := 0;
{$IFEND}
    while I < Length(str) do
    begin
{$IF defined(ANDROID) or defined(IOS)}
      str2 := str2 + str.Chars[Length(str) - I - 1];
      str2 := str2 + str.Chars[Length(str) - I];
{$ELSE}
      str2 := str2 + str[Length(str) - I - 1];
      str2 := str2 + str[Length(str) - I];
{$IFEND}
      I := I + 2;
    end;
    Result := HexaToBase32(str2);
  end;
end;

function TConvert.KeyTRSAEncSignToKeyRSAOpenSSL(strKey: string): string;
begin
  Result := KeyRSAOpenSSLToKeyTRSAEncSign(strKey);
end;

function TConvert.Base58Encode(const value: uint64): string;
var
  quotient, remainder, tmp: uint64;
begin
  Result := '';
  tmp := value;
  while (tmp >= 58) do
  begin
    quotient := tmp div 58;
    remainder := tmp mod 58;
    Result := Copy(BASE58_ALPHABET, remainder + 1, 1) + Result;
    tmp := quotient;
  end;
  Result := Copy(BASE58_ALPHABET, tmp + 1, 1) + Result;
end;

function TConvert.Base58Decode(const encoded: string): uint64;
var
  I: Integer;
  index: uint64;
  base: uint64;
begin
  Result := 0;
  base := 1;
  for I := Length(encoded) - 1 downto 0 do
  begin
    index := Pos(Copy(encoded, I + 1, 1), BASE58_ALPHABET) - 1;
    Result := Result + (index * base);
    base := 58 * base;
  end;
end;

function TConvert.FormatToString(str: string): string;
var
  hexastr, copystr: string;
  I: Integer;
begin
  if AType = raw then
  begin
    Result := str;
    Exit;
  end
  else if AType = base64 then
    hexastr := Base64ToHexa(str)
  else if AType = base64url then
    hexastr := Base64urlToHexa(str)
  else if AType = base32 then
    hexastr := Base32ToHexa(str)
  else if AType = hexa then
    hexastr := str;
  Result := '';
  I := 1;
  while I < Length(hexastr) do
  begin
    copystr := Copy(hexastr, I, 2);
    Result := Result + Char(StrToInt('$' + copystr));
    I := I + 2;
  end;
end;

function TConvert.StringToFormat(charstring: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  y: PAnsiChar;
  J: string;
{$ELSE}
  y: PByte;
{$IFEND}
  I: Integer;

begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  y := GetMemory(Length(charstring) * SizeOf(AnsiChar));
  for I := 0 to Length(charstring) - 1 do
  begin
    if Ord(charstring[I + 1]) < 256 then
      J := IntToHex(Ord(charstring[I + 1]), 1)
    else
      J := IntToHex(Ord(AnsiString(charstring)[I + 1]), 1);
    y[I] := AnsiChar(StrToInt('$' + J));
  end;
{$ELSE}
  y := GetMemory(Length(charstring) * SizeOf(Byte));
{$IF defined(LINUX)}
  for I := 1 to Length(charstring) do
    y[I - 1] := Byte(charstring[I]);
{$ELSE}
  for I := 0 to Length(charstring) - 1 do
    y[I] := Byte(charstring.Chars[I]);
{$IFEND}
{$IFEND}
  try
    Result := CharToFormat(y, Length(charstring));
  finally
    FreeMem(y);
  end;

end;

function TConvert.Base32ToHexa(base32String: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;
  AType := base32;
  charlen := Length(base32String);
  while charlen mod 8 <> 0 do
  begin
    base32String := base32String + '=';
    charlen := Length(base32String);
  end;
  charlen := CharLength(base32String);
  charstring := FormatToChar(base32String);
  try
    AType := hexa;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.HexaToBase32(hexaString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;
  AType := hexa;
  charlen := CharLength(hexaString);
  charstring := FormatToChar(hexaString);
  try
    AType := base32;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base32ToBase64url(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base32;
  charlen := Length(inString);
  while charlen mod 8 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base64url;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base64urlToBase32(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base64url;
  charlen := Length(inString);
  while charlen mod 4 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base32;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base32ToBase64(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base32;
  charlen := Length(inString);
  while charlen mod 8 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base64;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;

end;

function TConvert.Base64ToBase32(inString: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  charstring: PAnsiChar;
{$ELSE}
  charstring: PByte;
{$IFEND}
  charlen: Integer;
  saveType: TConvertType;
begin
  saveType := AType;

  AType := base64;
  charlen := Length(inString);
  while charlen mod 4 <> 0 do
  begin
    inString := inString + '=';
    charlen := Length(inString);
  end;
  charlen := CharLength(inString);
  charstring := FormatToChar(inString);
  try
    AType := base32;
    Result := CharToFormat(charstring, charlen);
  finally
    FreeMem(charstring);
    AType := saveType;
  end;
end;

function TConvert.TBytesToString(const t: TBytes): string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Length(t) - 1 do
    Result := Result + Char(t[I]);
end;

function TConvert.StringToTBytes(const str: string): TBytes;
var
  I, c, test: Integer;
begin
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
  c := 1;
{$ELSE}
  c := 0;
{$IFEND}
  test := TEncoding.UTF8.GetByteCount(str);
  if test <> Length(str) then
    raise ECryptoPack.Create
      ('Misuse of this function: Unicode characters are not accepted!');
  SetLength(Result, Length(str));
  for I := 0 to Length(str) - 1 do
{$IF (defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS))}
    Result[I] := Byte(str[I + c]);
{$ELSE}
    Result[I] := Byte(str.Chars[I + c]);
{$IFEND}
end;

function TConvert.StringToByteArray(str: string): TArray<Byte>;
var
  s: string;
  I: Integer;
begin
  s := str;
  SetLength(Result, Length(s));
  for I := 0 to Length(s) - 1 do
{$IF not((defined(MSWINDOWS) or defined(MACOS) or defined(LINUX)) and (not defined(IOS)))}
    Result[I] := Ord(s.Chars[I]);
{$ELSE}
    if Ord(s[I + 1]) < 256 then
      Result[I] := Ord(s[I + 1])
    else
      Result[I] := Ord(RawByteString(s)[I + 1]);
{$IFEND}
end;

function TConvert.UnicodeStringToFormat(str: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  Buffer: PAnsiChar;
{$ELSE}
  Buffer: PByte;
{$IFEND}
begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  Buffer := UnicodeToPAnsiChar(str);
  try
    Result := CharToFormat(Buffer, PAnsiCharFromUnicodeLength(str));
  finally
    FreeMem(Buffer);
  end;
{$ELSE}
  Buffer := PByte(TEncoding.UTF8.GetBytes(str));
  Result := CharToFormat(Buffer, TEncoding.UTF8.GetByteCount(str));
{$IFEND}
end;

function TConvert.FormatToUnicodeString(str: string): string;
var
  s: string;
begin
  s := FormatToString(str);
  Result := StringToUnicode(s);
end;

function TConvert.RandomString(len: Integer): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  Buffer: PAnsiChar;
{$ELSE}
  Buffer: PByte;
{$IFEND}
  err: Integer;
begin
  Buffer := GetMemory(len + 1);
  try
    err := RandomBuffer(len, Buffer);
    if err <> 0 then
      raise ECryptoPack.Create(err);

    Result := CharToFormat(Buffer, len);
  finally
    FreeMem(Buffer);
  end;
end;

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

function TConvert.StringToBuffer(str: string; u: TUnicode; var msgLen: Integer)
  : PAnsiChar;
{$ELSE}

function TConvert.StringToBuffer(str: string; u: TUnicode;
  var msgLen: Integer): PByte;
var
  t: TArray<Byte>;
{$IFEND}
begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  if u = yesUni then
  begin
    Result := UnicodeToPAnsiChar(str);
    msgLen := PAnsiCharFromUnicodeLength(str);
  end
  else
  begin
    Result := StringToPAnsiChar(str);
    msgLen := Length(str);
  end;

{$ELSE}
  if Length(str) > 0 then
  begin
    if u = yesUni then
    begin
      msgLen := TEncoding.UTF8.GetByteCount(str);
      t := TEncoding.UTF8.GetBytes(str);
    end
    else
    begin
      msgLen := Length(str);
      t := StringToByteArray(str);
    end;
    Result := GetMemory((Length(t) + 1) * SizeOf(Byte));
    Move(t[Low(t)], Result^, Length(t));
    Result[High(t) + 1] := Byte(#0);
    if u = noUni then
      StringToPByte(str, Result);
  end
  else
  begin
    Result := GetMemory(SizeOf(Byte));
    Result[0] := Byte(#0);
    msgLen := 1;
  end;
{$IFEND}
end;

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

function TConvert.StringToBufferA(str: string; u: TUnicode): PAnsiChar;
{$ELSE}

function TConvert.StringToBufferA(str: string; u: TUnicode): PByte;
var
  t: TArray<Byte>;
{$IFEND}
begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  if u = yesUni then
    Result := UnicodeToPAnsiChar(str)
  else
    Result := StringToPAnsiChar(str);

{$ELSE}
  if Length(str) > 0 then
  begin
    if u = yesUni then
      t := TEncoding.UTF8.GetBytes(str)
    else
      t := StringToByteArray(str);
    Result := GetMemory((Length(t) + 1) * SizeOf(Byte));
    Move(t[Low(t)], Result^, Length(t));
    Result[High(t) + 1] := Byte(#0);
    if u = noUni then
      StringToPByte(str, Result);
  end
  else
  begin
    Result := GetMemory(SizeOf(Byte));
    Result[0] := Byte(#0);
  end;
{$IFEND}
end;

function PosExZ(const SubString: string; const s: string;
  const StartIndex: Integer): Integer;
begin
  Result := Pos(SubString, Copy(s, StartIndex, Length(s) - StartIndex));
  if Result > 0 then
    Result := Result + StartIndex;
end;

{$IF (defined(WIN32) or defined(MACOS32)) and (not defined(IOS))}

procedure __llmul;
asm
  jmp System.@_llmul
end;

procedure __llumod;
asm
  jmp System.@_llumod
end;

procedure __lludiv;
asm
  jmp System.@_lludiv
end;

procedure __lldiv;
asm
  jmp System.@_lldiv
end;

procedure __aulldiv; external 'ntdll' name '_aulldiv';

procedure __alldiv; external 'ntdll' name '_alldiv';

procedure __aullrem; external 'ntdll' name '_aullrem';

procedure __chkstk; external 'ntdll' name '_chkstk';

function _wcscpy(destination: PChar; source: PChar): PChar;
  external msvcrt name 'wcscpy';

{$IFEND}
{$IF defined(WIN64)}
procedure __chkstk; external 'ntdll';
{$IFEND}
{$IF defined(WIN32)}
procedure __chkstk_noalloc; external 'ntdll';
{$IFEND}

end.
