{ *************************************************************************** }
{ TMS Cryptography Pack                                                       }
{ for Delphi & C++Builder                                                     }
{                                                                             }
{ Written by Cyberens Technologies & Services                                 }
{ Exclusively commercialized by tmssoftware.com bvba                          }
{ copyright  2016 - 2020 by Cyberens Technologies & Services                 }
{ Email : info@tmssoftware.com                                                }
{ Web : http://www.tmssoftware.com                                            }
{                                                                             }
{ The source code is given as is. The author is not responsible               }
{ for any possible damage done due to the use of this code.                   }
{ The component can be freely used in any application. The complete           }
{ source code remains property of the author and may not be distributed,      }
{ published, given or sold in any form as such. No parts of the source        }
{ code can be included in any other component or application without          }
{ written authorization of the author.                                        }
{ *************************************************************************** }

unit SPECKObj;

interface

{$I TMSDEFS.INC}
{$I TMSCRYPTO.INC}

uses
  CryptBase, MiscObj, Classes;

{$POINTERMATH ON}

type
  TSPECKWordSizeBits = (wsb16, wsb24, wsb32, wsb48, wsb64);
  TSPECKKeySizeWords = (ksw2, ksw3, ksw4);
  TSPECKType = (stECB, stCBC, stOFB);
  TSPECKIVMode = (rand, userdefined);
  TSPECKPaddingMode = (PKCS7, nopadding);
  TSPECKKey16 = PWord;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  TSPECKKey24 = PPAnsiChar;
{$ELSE}
  PPByte = ^PByte;
  TSPECKKey24 = PPByte;
{$IFEND}
  TSPECKKey32 = PInteger;
  PPWord = ^PWord;
  TSPECKKey48 = PPWord;
  TSPECKKey64 = PInt64;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  TSPECKBlock = PAnsiChar;
{$ELSE}
  TSPECKBlock = PByte;
{$IFEND}

  TSPECKEncryption = class(TTMSCryptBase)
  private
    FWordSizeBitsNum: Integer;
    FKeySizeWordsNum: Integer;
    FWordSizeBits: TSPECKWordSizeBits;
    FKeySizeWords: TSPECKKeySizeWords;
    FKey: string;
    FType: TSPECKType;
    FOutputFormat: TConvertType;
    FIV: string;
    FUni: TUnicode;
    FIVMode: TSPECKIVMode;
    FPaddingMode: TSPECKPaddingMode;
    FProgress: Integer;
    FOnChange: TNotifyEvent;
    procedure SetKey(const Value: string);
    procedure SetWordSizeBits(const Value: TSPECKWordSizeBits);
    procedure SetKeySizeWords(const Value: TSPECKKeySizeWords);
    procedure SetIV(const Value: string);
    procedure SetProgress(P: Integer);
    procedure SPECKEncryptionLoop16(P: TSPECKBlock; c: TSPECKBlock;
      K16: TSPECKKey16; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKEncryptionLoop24(P: TSPECKBlock; c: TSPECKBlock;
      K24: TSPECKKey24; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKEncryptionLoop32(P: TSPECKBlock; c: TSPECKBlock;
      K32: TSPECKKey32; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKEncryptionLoop48(P: TSPECKBlock; c: TSPECKBlock;
      K48: TSPECKKey48; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKEncryptionLoop64(P: TSPECKBlock; c: TSPECKBlock;
      K64: TSPECKKey64; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKDecryptionLoop16(c: TSPECKBlock; d: TSPECKBlock;
      K16: TSPECKKey16; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKDecryptionLoop24(c: TSPECKBlock; d: TSPECKBlock;
      K24: TSPECKKey24; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKDecryptionLoop32(c: TSPECKBlock; d: TSPECKBlock;
      K32: TSPECKKey32; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKDecryptionLoop48(c: TSPECKBlock; d: TSPECKBlock;
      K48: TSPECKKey48; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);
    procedure SPECKDecryptionLoop64(c: TSPECKBlock; d: TSPECKBlock;
      K64: TSPECKKey64; keySizeWordsN: Integer; AType: TSPECKType;
      IV: TSPECKBlock);

  public
    Constructor Create(AOwner: TComponent); overload; override;
    Constructor Create; reintroduce; overload;
    Constructor Create(wordSizeBits: TSPECKWordSizeBits;
      keySizeWords: TSPECKKeySizeWords; key: string; AType: TSPECKType;
      OutputFormat: TConvertType; paddingMode: TSPECKPaddingMode;
      uni: TUnicode); reintroduce; overload;
    Constructor Create(wordSizeBits: TSPECKWordSizeBits;
      keySizeWords: TSPECKKeySizeWords; key: string; AType: TSPECKType;
      OutputFormat: TConvertType; paddingMode: TSPECKPaddingMode; uni: TUnicode;
      IV: string); reintroduce; overload;
    Destructor Destroy; override;
    function Encrypt(s: string): string;
    function Decrypt(s: string): string;
    procedure EncryptFileW(s, o: string);
    procedure DecryptFileW(s, o: string);
    procedure EncryptStream(s: TStream; var o: TStream);
    procedure DecryptStream(s: TStream; var o: TStream);
  published
    property wordSizeBits: TSPECKWordSizeBits read FWordSizeBits
      write SetWordSizeBits default wsb32;
    property keySizeWords: TSPECKKeySizeWords read FKeySizeWords
      write SetKeySizeWords default ksw4;
    property key: string read FKey write SetKey;
    property AType: TSPECKType read FType write FType default stCBC;
    property OutputFormat: TConvertType read FOutputFormat write FOutputFormat
      default hexa;
    property IVMode: TSPECKIVMode read FIVMode write FIVMode default rand;
    property IV: string read FIV write SetIV;
    property paddingMode: TSPECKPaddingMode read FPaddingMode write FPaddingMode
      default PKCS7;
    property Unicode: TUnicode read FUni write FUni default yesUni;
    property Progress: Integer read FProgress write SetProgress default 0;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TSPECKEncryption }

{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}

function SPECK_EncryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
function SPECK_DecryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
function SPECK_EncryptString(inputString: PAnsiChar; outputBuffer: PAnsiChar;
  outputSizeBytes: PInteger; key: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl;
function SPECK_EncryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer)
  : Integer; cdecl;
function SPECK_DecryptString(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  outputString: PAnsiChar; ecbKey: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl;
function SPECK_DecryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer)
  : Integer; cdecl;

function SPECK16_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey16): Integer; cdecl;
function SPECK24_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey24): Integer; cdecl;
function SPECK32_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey32): Integer; cdecl;
function SPECK48_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey48): Integer; cdecl;
function SPECK64_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey64): Integer; cdecl;
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer; cdecl;
function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
{$ELSE}
function SPECK_EncryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
function SPECK_DecryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
function SPECK_EncryptString(inputString: PByte; outputBuffer: PByte;
  outputSizeBytes: PInteger; key: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer;
function SPECK_EncryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer;
function SPECK_DecryptString(inputBuffer: PByte; inputSizeBytes: Integer;
  outputString: PByte; ecbKey: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer;
function SPECK_DecryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer;

function SPECK16_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey16): Integer;
function SPECK24_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey24): Integer;
function SPECK32_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey32): Integer;
function SPECK48_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey48): Integer;
function SPECK64_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey64): Integer;
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer;
function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;

{$IFEND}

implementation

uses Math, SysUtils, DateUtils,
{$IFDEF MSWINDOWS}
  Winapi.Windows, AnsiStrings, crtl,
{$ENDIF MSWINDOWS}
{$IF defined(MACOS) and (not defined(IOS))}
  AnsiStrings,
{$IFEND}
{$IFDEF POSIX}
  Posix.Dlfcn, Posix.Fcntl, Posix.SysStat, Posix.SysTime, Posix.SysTypes,
  Posix.Unistd,
{$ENDIF POSIX}
{$IFDEF PC_MAPPED_EXCEPTIONS}
  System.Internal.Unwinder,
{$ENDIF PC_MAPPED_EXCEPTIONS}
  System.SysConst;

const
  ADVAPI32 = 'advapi32.dll';
  msvcrt = 'msvcrt.dll';
{$IFNDEF WINXP}
  BCRYPT = 'Bcrypt.dll';
{$ENDIF}
{$IF not defined(IDEVERSION1021)}
  RANDOMDLL = 'RandomDLL.dll'; // Copy the DLL from Win64 directory
  // into C:\Windows\SysWOW64 if you are running 32 bit Windows
  // or into C:\Windows\System32 if you are running 64 bit Windows
{$IFEND}
  libc = '/usr/lib/libc.dylib';
  PROV_RSA_FULL = 1;
  CRYPT_VERIFYCONTEXT = $F0000000;
{$IF (defined(WIN32) or defined(MACOS)) and (not defined(IOS))}
  PU = '_';
{$ELSE}
  PU = '';
{$IFEND}
{$IF defined(WIN32)}
{$LINK './libWIN32/SPECK_mode.obj'}
{$LINK './libWIN32/SPECK.obj'}
{$IFDEF WINXP}
{$LINK './libWIN32/elt_rep_XP.obj'}
{$ELSE}
{$LINK './libWIN32/elt_rep.obj'}
{$ENDIF}
{$LINK './libWIN32/error.obj'}
{$ELSEIF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
{$LINK './libWIN64/SPECK_mode.o'}
{$LINK './libWIN64/SPECK.o'}
{$LINK './libWIN64/elt_rep.o'}
{$LINK './libWIN64/error.o'}
{$ELSE}
{$LINK './libWIN64/SPECK_mode.obj'}
{$LINK './libWIN64/SPECK.obj'}
{$LINK './libWIN64/elt_rep.obj'}
{$LINK './libWIN64/error.obj'}
{$IFEND}
{$ELSEIF defined(MACOS32) and not defined(IOS)}
{$LINK './libOSX32/SPECK_mode.o'}
{$LINK './libOSX32/SPECK.o'}
{$LINK './libOSX32/elt_rep.o'}
{$LINK './libOSX32/error.o'}
{$IFEND}
{$IF (defined(MSWINDOWS) or defined(MACOS32)) and (not defined(IOS))}
function SPECK_EncryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
  external name PU + 'SPECK_EncryptFile';
function SPECK_DecryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
  external name PU + 'SPECK_DecryptFile';
function SPECK_EncryptString(inputString: PAnsiChar; outputBuffer: PAnsiChar;
  outputSizeBytes: PInteger; key: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl; external name PU + 'SPECK_EncryptString';
function SPECK_EncryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer): Integer;
  cdecl; external name PU + 'SPECK_EncryptBinaryBuffer';
function SPECK_DecryptString(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  outputString: PAnsiChar; ecbKey: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl; external name PU + 'SPECK_DecryptString';
function SPECK_DecryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer): Integer;
  cdecl; external name PU + 'SPECK_DecryptBinaryBuffer';

function SPECK16_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey16): Integer; cdecl; external name PU + 'SPECK16_KeyExpansion';
function SPECK24_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey24): Integer; cdecl; external name PU + 'SPECK24_KeyExpansion';
function SPECK32_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey32): Integer; cdecl; external name PU + 'SPECK32_KeyExpansion';
function SPECK48_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey48): Integer; cdecl; external name PU + 'SPECK48_KeyExpansion';
function SPECK64_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey64): Integer; cdecl; external name PU + 'SPECK64_KeyExpansion';
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer; cdecl;
  external name PU + 'Padding_speckmode';
function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK16_Encryption';
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK16_Decryption';
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK24_Encryption';
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK24_Decryption';
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK32_Encryption';
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK32_Decryption';
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK48_Encryption';
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK48_Decryption';
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK64_Encryption';
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external name PU + 'SPECK64_Decryption';

function CharToHexa(MyChar: PAnsiChar; LenChar: Integer; HexaChar: PAnsiChar)
  : Integer; cdecl; external name PU + 'CharToHexa';
function HexaToChar(HexaChar: PAnsiChar; MyChar: PAnsiChar): Integer; cdecl;
  external name PU + 'HexaToChar';

{$IF defined(WIN32)}
{$IFDEF WINXP}
function CryptAcquireContextW(phProv: pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
{$ELSE}
function BCryptOpenAlgorithmProvider(out phAlgorithm: THandle;
  pszAlgId: LPCWSTR; pszProvider: LPCWSTR; dwProvType: DWORD; dwFlags: DWORD)
  : LongInt; stdcall; external BCRYPT name 'BCryptOpenAlgorithmProvider';
function BCryptCloseAlgorithmProvider(hAlgorithm: THandle; dwFlags: ULONG)
  : LongInt; stdcall; external BCRYPT name 'BCryptCloseAlgorithmProvider';
function BCryptGenRandom(hAlgorithm: THandle; out pbBuffer: PUCHAR;
  cbBuffer, dwFlags: ULONG): LongInt; stdcall;
  external BCRYPT name 'BCryptGenRandom';
{$ENDIF}
function _fopen(Str1: PByte; Str2: PByte): pointer; cdecl;
  external msvcrt name 'fopen';

function _fread(Ptr1: pointer; Int1: Integer; Int2: Integer; Ptr2: pointer)
  : Integer; cdecl; external msvcrt name 'fread';

function _fseek(Ptr1: pointer; Int1: Integer; Int2: Integer): Integer; cdecl;
  external msvcrt name 'fseek';

function _ftell(Ptr: pointer): Integer; cdecl; external msvcrt name 'ftell';

function _fwrite(Ptr1: pointer; Int1: Integer; Int2: Integer; Ptr2: pointer)
  : Integer; cdecl; external msvcrt name 'fwrite';
function _fclose(Ptr: pointer): Integer; cdecl; external msvcrt name 'fclose';
{$IFEND}
{$IF defined(WIN32) or defined(MACOS32)}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomBuffer';
function RandomInt(): Integer; cdecl; external name PU + 'RandomInt';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer; cdecl;
  external name PU + 'RandomUBuffer';
function RandomUInt(): Integer; cdecl; external name PU + 'RandomUInt';

{$IFEND}
{$IF defined(MACOS32)}
function malloc(size: size_t): pointer; cdecl; external libc name PU + 'malloc';
procedure _memcpy(dest, source: pointer; count: Integer); cdecl;
  external libc name PU + 'memcpy';
procedure free(P: pointer); cdecl; external libc name PU + 'free';
function realloc(P: pointer; NewSize: size_t): pointer; cdecl;
  external libc name PU + 'realloc';
function _open(PathName: MarshaledAString; Flags: Integer): Integer; cdecl;
  varargs; external libc name PU + 'open';
function _close(filedes: Integer): Integer; cdecl;
  external libc name PU + 'close';
function _lseek(filedes: Integer; offset: off_t; whence: Integer): off_t; cdecl;
  external libc name PU + 'lseek';
function _read(Handle: Integer; Buffer: pointer; count: size_t): ssize_t; cdecl;
  external libc name PU + 'read';
function _strlen(S1: MarshaledAString): size_t; cdecl;
  external libc name PU + 'strlen';
function _malloc(size: size_t): pointer; cdecl;
  external libc name PU + 'malloc';
procedure _free(P: pointer); cdecl; external libc name PU + 'free';
function _realloc(P: pointer; NewSize: size_t): pointer; cdecl;
  external libc name PU + 'realloc';
function _memset(var b; c: Integer; len: size_t): pointer; cdecl;
  external libc name PU + 'memset';
function _strcat(S1: MarshaledAString; s2: MarshaledAString): MarshaledAString;
  cdecl; external libc name PU + 'strcat';
function _write(Handle: Integer; Buffer: pointer; count: size_t): ssize_t;
  cdecl; external libc name PU + 'write';

{$IFEND}
{$IF defined(WIN64)}
{$IF defined(IDEVERSION1021)}
function CryptAcquireContextW(phProv: pointer; pszContainer: LPCSTR;
  pszProvider: LPCSTR; dwProvType: DWORD; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptAcquireContextW';
function CryptReleaseContext(hProv: pointer; dwFlags: DWORD): BOOL; stdcall;
  external ADVAPI32 name 'CryptReleaseContext';
function CryptGenRandom(hProv: ULONG; dwLen: DWORD; pbBuffer: PByte): BOOL;
  stdcall; external ADVAPI32 name 'CryptGenRandom';
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomBuffer';
function RandomInt(): Integer; external name PU + 'RandomInt';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external name PU + 'RandomUBuffer';
function RandomUInt(): Integer; external name PU + 'RandomUInt';
{$ELSE}
function RandomBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external RANDOMDLL name 'RandomBuffer';
function RandomUBuffer(len: Integer; MyBuffer: PAnsiChar): Integer;
  external RANDOMDLL name 'RandomUBuffer';
function RandomInt(): Integer; external RANDOMDLL name 'RandomInt';
function RandomUInt(): Integer; external RANDOMDLL name 'RandomUInt';
{$IFEND}
function fclose(Ptr: pointer): Integer; external msvcrt name 'fclose';

function fopen(Str1: PAnsiChar; Str2: PAnsiChar): pointer;
  external msvcrt name 'fopen';

function fread(Ptr1: pointer; Int1: Integer; Int2: Integer; Ptr2: pointer)
  : Integer; external msvcrt name 'fread';

function fseek(Ptr1: pointer; Int1: Integer; Int2: Integer): Integer;
  external msvcrt name 'fseek';

function ftell(Ptr: pointer): Integer; external msvcrt name 'ftell';

function fwrite(Ptr1: pointer; Int1: Integer; Int2: Integer; Ptr2: pointer)
  : Integer; external msvcrt name 'fwrite';
{$IFEND}
{$ELSEIF defined(ANDROID) or defined(IOS)}
function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomInt(): Integer; external 'libTMSCPLib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPLib.a';
function RandomUInt(): Integer; external 'libTMSCPLib.a';

function SPECK_EncryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
  external 'libTMSCPLib.a';
function SPECK_DecryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
  external 'libTMSCPLib.a';
function SPECK_EncryptString(inputString: PByte; outputBuffer: PByte;
  outputSizeBytes: PInteger; key: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer; external 'libTMSCPLib.a';
function SPECK_EncryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer; external 'libTMSCPLib.a';
function SPECK_DecryptString(inputBuffer: PByte; inputSizeBytes: Integer;
  outputString: PByte; ecbKey: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer; external 'libTMSCPLib.a';
function SPECK_DecryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer; external 'libTMSCPLib.a';

function SPECK16_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey16): Integer; external 'libTMSCPLib.a';
function SPECK24_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey24): Integer; external 'libTMSCPLib.a';
function SPECK32_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey32): Integer; external 'libTMSCPLib.a';
function SPECK48_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey48): Integer; external 'libTMSCPLib.a';
function SPECK64_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey64): Integer; external 'libTMSCPLib.a';
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer;
  external 'libTMSCPLib.a';

function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPLib.a';

function CharToHexa(MyChar: PByte; LenChar: Integer; HexaChar: PByte): Integer;
  external 'libTMSCPLib.a';
function HexaToChar(HexaChar: PByte; MyChar: PByte): Integer;
  external 'libTMSCPLib.a';
{$ELSEIF defined(LINUX)}
function RandomBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomInt(): Integer; external 'libTMSCPlib.a';
function RandomUBuffer(len: Integer; MyBuffer: PByte): Integer;
  external 'libTMSCPlib.a';
function RandomUInt(): Integer; external 'libTMSCPlib.a';

function SPECK_EncryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
  external 'libTMSCPlib.a';
function SPECK_DecryptFile(inputFilePath: PByte; outputFilePath: PByte;
  key: PByte; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PByte; padding: Integer; mode: Integer): Integer;
  external 'libTMSCPlib.a';
function SPECK_EncryptString(inputString: PByte; outputBuffer: PByte;
  outputSizeBytes: PInteger; key: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer; external 'libTMSCPlib.a';
function SPECK_EncryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer; external 'libTMSCPlib.a';
function SPECK_DecryptString(inputBuffer: PByte; inputSizeBytes: Integer;
  outputString: PByte; ecbKey: PByte; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PByte; padding: Integer;
  mode: Integer): Integer; external 'libTMSCPlib.a';
function SPECK_DecryptBinaryBuffer(inputBuffer: PByte; inputSizeBytes: Integer;
  outputBuffer: PByte; outputSizeBytes: PInteger; ecbKey: PByte;
  wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer; IV: PByte;
  padding: Integer; mode: Integer): Integer; external 'libTMSCPlib.a';

function SPECK16_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey16): Integer; external 'libTMSCPlib.a';
function SPECK24_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey24): Integer; external 'libTMSCPlib.a';
function SPECK32_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey32): Integer; external 'libTMSCPlib.a';
function SPECK48_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey48): Integer; external 'libTMSCPlib.a';
function SPECK64_KeyExpansion(keywords: PByte; keySizeWords: Integer;
  K: TSPECKKey64): Integer; external 'libTMSCPlib.a';
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer;
  external 'libTMSCPlib.a';

function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; var outputBuffer: TSPECKBlock): Integer;
  external 'libTMSCPlib.a';

function CharToHexa(MyChar: PByte; LenChar: Integer; HexaChar: PByte): Integer;
  external 'libTMSCPlib.a';
function HexaToChar(HexaChar: PByte; MyChar: PByte): Integer;
  external 'libTMSCPlib.a';
{$ELSE}
function SPECK_EncryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
  external object'./libOSX64/SPECK_mode.o' name 'SPECK_EncryptFile';
function SPECK_DecryptFile(inputFilePath: PAnsiChar; outputFilePath: PAnsiChar;
  key: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer; IVIn: Integer;
  IV: PAnsiChar; padding: Integer; mode: Integer): Integer; cdecl;
  external object'./libOSX64/SPECK_mode.o' name 'SPECK_DecryptFile';
function SPECK_EncryptString(inputString: PAnsiChar; outputBuffer: PAnsiChar;
  outputSizeBytes: PInteger; key: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl;
  external object'./libOSX64/SPECK_mode.o' name 'SPECK_EncryptString';
function SPECK_EncryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer): Integer;
  cdecl; external object'./libOSX64/SPECK_mode.o' name
  'SPECK_EncryptBinaryBuffer';
function SPECK_DecryptString(inputBuffer: PAnsiChar; inputSizeBytes: Integer;
  outputString: PAnsiChar; ecbKey: PAnsiChar; wordSizeBits: Integer;
  keySizeWords: Integer; IVIn: Integer; IV: PAnsiChar; padding: Integer;
  mode: Integer): Integer; cdecl;
  external object'./libOSX64/SPECK_mode.o' name 'SPECK_DecryptString';
function SPECK_DecryptBinaryBuffer(inputBuffer: PAnsiChar;
  inputSizeBytes: Integer; outputBuffer: PAnsiChar; outputSizeBytes: PInteger;
  ecbKey: PAnsiChar; wordSizeBits: Integer; keySizeWords: Integer;
  IVIn: Integer; IV: PAnsiChar; padding: Integer; mode: Integer): Integer;
  cdecl; external object'./libOSX64/SPECK_mode.o' name
  'SPECK_DecryptBinaryBuffer';

function SPECK16_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey16): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK16_KeyExpansion';
function SPECK24_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey24): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK24_KeyExpansion';
function SPECK32_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey32): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK32_KeyExpansion';
function SPECK48_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey48): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK48_KeyExpansion';
function SPECK64_KeyExpansion(keywords: PAnsiChar; keySizeWords: Integer;
  K: TSPECKKey64): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK64_KeyExpansion';
function Padding_speckmode(P: TSPECKBlock; PLength, i: Integer): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'Padding_speckmode';
function SPECK16_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK16_Encryption';
function SPECK16_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey16;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK16_Decryption';
function SPECK24_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK24_Encryption';
function SPECK24_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey24;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK24_Decryption';
function SPECK32_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK32_Encryption';
function SPECK32_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey32;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK32_Decryption';
function SPECK48_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK48_Encryption';
function SPECK48_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey48;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK48_Decryption';
function SPECK64_Encryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK64_Encryption';
function SPECK64_Decryption(inputBuffer: TSPECKBlock; K: TSPECKKey64;
  keySizeWordsN: Integer; outputBuffer: TSPECKBlock): Integer; cdecl;
  external object'./libOSX64/SPECK.o' name 'SPECK64_Decryption';

function CharToHexa(MyChar: PAnsiChar; LenChar: Integer; HexaChar: PAnsiChar)
  : Integer; cdecl; external object'./libOSX64/elt_rep.o' name 'CharToHexa';
function HexaToChar(HexaChar: PAnsiChar; MyChar: PAnsiChar): Integer; cdecl;
  external object'./libOSX64/elt_rep.o' name 'HexaToChar';
{$IFEND}

constructor TSPECKEncryption.Create(AOwner: TComponent);
begin
  self.SetWordSizeBits(wsb32);
  self.SetKeySizeWords(ksw4);
  self.FType := stCBC;
  self.FOutputFormat := hexa;
  self.FPaddingMode := PKCS7;
  self.FIVMode := rand;
  self.FUni := yesUni;
  self.FOnChange := nil;
  inherited Create(AOwner);
end;

constructor TSPECKEncryption.Create;
begin
  self.SetWordSizeBits(wsb32);
  self.SetKeySizeWords(ksw4);
  self.FType := stCBC;
  self.FOutputFormat := hexa;
  self.FPaddingMode := PKCS7;
  self.FIVMode := rand;
  self.FUni := yesUni;
  self.FOnChange := nil;
end;

constructor TSPECKEncryption.Create(wordSizeBits: TSPECKWordSizeBits;
  keySizeWords: TSPECKKeySizeWords; key: string; AType: TSPECKType;
  OutputFormat: TConvertType; paddingMode: TSPECKPaddingMode; uni: TUnicode);
begin
  self.SetWordSizeBits(wordSizeBits);
  self.SetKeySizeWords(keySizeWords);
  self.SetKey(key);
  self.FType := AType;
  self.FOutputFormat := OutputFormat;
  self.FPaddingMode := paddingMode;
  self.FIVMode := rand;
  self.FUni := uni;
  self.FOnChange := nil;
end;

constructor TSPECKEncryption.Create(wordSizeBits: TSPECKWordSizeBits;
  keySizeWords: TSPECKKeySizeWords; key: string; AType: TSPECKType;
  OutputFormat: TConvertType; paddingMode: TSPECKPaddingMode; uni: TUnicode;
  IV: string);
begin
  self.SetWordSizeBits(wordSizeBits);
  self.SetKeySizeWords(keySizeWords);
  self.SetKey(key);
  self.FType := AType;
  self.FOutputFormat := OutputFormat;
  self.FPaddingMode := paddingMode;
  self.FIVMode := userdefined;
  self.SetIV(IV);
  self.FUni := uni;
  self.FOnChange := nil;
end;

destructor TSPECKEncryption.Destroy;
begin
  FKey := StringOfChar(#0, (FWordSizeBitsNum * FKeySizeWordsNum) div 8);
  inherited;
end;

function TSPECKEncryption.Decrypt(s: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  outputBuffer, inputBuffer, KeyPB, IV: PAnsiChar;
{$ELSE}
  outputBuffer, inputBuffer, KeyPB, IV: PByte;
  i: Integer;
{$IFEND}
  err, temp: Integer;
  outputSizeBytes: PInteger;
  P: pointer;
  conv: TConvert;
  padd, IVIn: Boolean;
begin
  conv := TConvert.Create(OutputFormat);
  case FPaddingMode of
    PKCS7:
      begin
        temp := 0;
      end;
    nopadding:
      begin
        case FIVMode of
          rand:
            begin
              temp := conv.CharLength(s) - FWordSizeBitsNum div 4;
            end;
          userdefined:
            begin
              temp := conv.CharLength(s);
            end;
        end;

      end;
  end;
  P := @temp;
  outputSizeBytes := PInteger(P);
  outputBuffer := GetMemory(conv.CharLength(s) + 1);
  KeyPB := conv.StringToBufferA(FKey, noUni);
{$IF not((defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS)))}
  outputBuffer[conv.CharLength(s)] := Byte(#0);
  for i := 0 to conv.CharLength(s) do
    outputBuffer[i] := Byte(#0);
{$IFEND}
  inputBuffer := conv.FormatToChar(s);
  if FIVMode = userdefined then
    IV := conv.StringToBufferA(FIV, noUni)
  else
    IV := nil;
  try

    padd := (FPaddingMode = TSPECKPaddingMode.PKCS7);
    IVIn := (FIVMode = TSPECKIVMode.rand);
    err := SPECK_DecryptBinaryBuffer(inputBuffer, conv.CharLength(s),
      outputBuffer, outputSizeBytes, KeyPB, FWordSizeBitsNum, FKeySizeWordsNum,
      Integer(IVIn), IV, Integer(padd), Integer(FType));

    if err <> 0 then
      raise ECryptoPack.Create(err);

    Result := conv.CharToFormat(outputBuffer, outputSizeBytes^);
    Result := conv.FormatToString(Result);

    if FUni = yesUni then
      Result := conv.StringToUnicode(Result);
  finally
    FreeMem(inputBuffer);
    conv.free;
    FreeMem(outputBuffer);
    FreeMem(KeyPB);
    if FIVMode = userdefined then
      FreeMem(IV);
  end;

end;

procedure TSPECKEncryption.EncryptFileW(s, o: string);
var
  stream1, stream2: TFileStream;
  f: Integer;
begin

  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    if FileExists(o) then
    begin
      System.SysUtils.DeleteFile(o);
      f := FileCreate(o);
      FileClose(f);
    end
    else
    begin
      f := FileCreate(o);
      FileClose(f);
    end;
    stream2 := TFileStream.Create(o, fmOpenReadWrite);
    try
      self.EncryptStream(stream1, TStream(stream2));
    finally
      stream2.free;
    end;
  finally
    stream1.free;
  end;

end;

procedure TSPECKEncryption.SPECKEncryptionLoop16(P: TSPECKBlock; c: TSPECKBlock;
  K16: TSPECKKey16; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    FillChar(c^, 4, 0);
    err := SPECK16_Encryption(P, K16, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    for i := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      P[i] := AnsiChar(Byte(IV[i]) xor Byte(P[i]));
{$ELSE}
      P[i] := IV[i] xor P[i];
{$IFEND}
    err := SPECK16_Encryption(P, K16, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 4 * SizeOf(Byte));
  end
  else
  begin
    err := SPECK16_Encryption(IV, K16, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 4 * SizeOf(Byte));
    for i := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      c[i] := AnsiChar(Byte(P[i]) xor Byte(c[i]));
{$ELSE}
      c[i] := P[i] xor c[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKEncryptionLoop24(P: TSPECKBlock; c: TSPECKBlock;
  K24: TSPECKKey24; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK24_Encryption(P, K24, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    for i := 0 to 5 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      P[i] := AnsiChar(Byte(IV[i]) xor Byte(P[i]));
{$ELSE}
      P[i] := IV[i] xor P[i];
{$IFEND}
    err := SPECK24_Encryption(P, K24, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 6 * SizeOf(Byte));
  end
  else
  begin
    err := SPECK24_Encryption(IV, K24, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 6 * SizeOf(Byte));
    for i := 0 to 5 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      c[i] := AnsiChar(Byte(c[i]) xor Byte(P[i]));
{$ELSE}
      c[i] := P[i] xor c[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKEncryptionLoop32(P: TSPECKBlock; c: TSPECKBlock;
  K32: TSPECKKey32; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK32_Encryption(P, K32, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    for i := 0 to 7 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      P[i] := AnsiChar(Byte(IV[i]) xor Byte(P[i]));
{$ELSE}
      P[i] := IV[i] xor P[i];
{$IFEND}
    err := SPECK32_Encryption(P, K32, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 8 * SizeOf(Byte));
  end
  else
  begin
    err := SPECK32_Encryption(IV, K32, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 8 * SizeOf(Byte));
    for i := 0 to 7 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      c[i] := AnsiChar(Byte(c[i]) xor Byte(P[i]));
{$ELSE}
      c[i] := c[i] xor P[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKEncryptionLoop48(P: TSPECKBlock; c: TSPECKBlock;
  K48: TSPECKKey48; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK48_Encryption(P, K48, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    for i := 0 to 11 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      P[i] := AnsiChar(Byte(IV[i]) xor Byte(P[i]));
{$ELSE}
      P[i] := IV[i] xor P[i];
{$IFEND}
    err := SPECK48_Encryption(P, K48, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 12 * SizeOf(Byte));
  end
  else
  begin
    err := SPECK48_Encryption(IV, K48, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 12 * SizeOf(Byte));
    for i := 0 to 11 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      c[i] := AnsiChar(Byte(c[i]) xor Byte(P[i]));
{$ELSE}
      c[i] := c[i] xor P[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKEncryptionLoop64(P: TSPECKBlock; c: TSPECKBlock;
  K64: TSPECKKey64; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK64_Encryption(P, K64, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    for i := 0 to 15 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      P[i] := AnsiChar(Byte(IV[i]) xor Byte(P[i]));
{$ELSE}
      P[i] := IV[i] xor P[i];
{$IFEND}
    err := SPECK64_Encryption(P, K64, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 16 * SizeOf(Byte));
  end
  else
  begin
    err := SPECK64_Encryption(IV, K64, keySizeWordsN, c);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(c[0], IV[0], 16 * SizeOf(Byte));
    for i := 0 to 15 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      c[i] := AnsiChar(Byte(c[i]) xor Byte(P[i]));
{$ELSE}
      c[i] := c[i] xor P[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKDecryptionLoop16(c: TSPECKBlock; d: TSPECKBlock;
  K16: TSPECKKey16; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK16_Decryption(c, K16, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    err := SPECK16_Decryption(c, K16, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    for i := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(IV[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := IV[i] xor d[i];
{$IFEND}
    Move(c[0], IV[0], 4 * SizeOf(Byte));

  end
  else
  begin
    err := SPECK16_Encryption(IV, K16, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(d[0], IV[0], 4 * SizeOf(Byte));
    for i := 0 to 3 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(c[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := c[i] xor d[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKDecryptionLoop24(c: TSPECKBlock; d: TSPECKBlock;
  K24: TSPECKKey24; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK24_Decryption(c, K24, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    err := SPECK24_Decryption(c, K24, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    for i := 0 to 5 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(IV[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := IV[i] xor d[i];
{$IFEND}
    Move(c[0], IV[0], 6 * SizeOf(Byte));

  end
  else
  begin
    err := SPECK24_Encryption(IV, K24, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(d[0], IV[0], 6 * SizeOf(Byte));
    for i := 0 to 5 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(c[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := c[i] xor d[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKDecryptionLoop32(c: TSPECKBlock; d: TSPECKBlock;
  K32: TSPECKKey32; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK32_Decryption(c, K32, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    err := SPECK32_Decryption(c, K32, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    for i := 0 to 7 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(IV[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := IV[i] xor d[i];
{$IFEND}
    Move(c[0], IV[0], 8 * SizeOf(Byte));

  end
  else
  begin
    err := SPECK32_Encryption(IV, K32, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(d[0], IV[0], 8 * SizeOf(Byte));
    for i := 0 to 7 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(c[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := c[i] xor d[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKDecryptionLoop48(c: TSPECKBlock; d: TSPECKBlock;
  K48: TSPECKKey48; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK48_Decryption(c, K48, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    err := SPECK48_Decryption(c, K48, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    for i := 0 to 11 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(IV[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := IV[i] xor d[i];
{$IFEND}
    Move(c[0], IV[0], 12 * SizeOf(Byte));

  end
  else
  begin
    err := SPECK48_Encryption(IV, K48, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(d[0], IV[0], 12 * SizeOf(Byte));
    for i := 0 to 11 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(c[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := c[i] xor d[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.SPECKDecryptionLoop64(c: TSPECKBlock; d: TSPECKBlock;
  K64: TSPECKKey64; keySizeWordsN: Integer; AType: TSPECKType; IV: TSPECKBlock);
var
  err, i: Integer;
begin
  if AType = TSPECKType.stECB then
  begin
    err := SPECK64_Decryption(c, K64, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
  end
  else if AType = TSPECKType.stCBC then
  begin
    err := SPECK64_Decryption(c, K64, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    for i := 0 to 15 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(IV[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := IV[i] xor d[i];
{$IFEND}
    Move(c[0], IV[0], 16 * SizeOf(Byte));

  end
  else
  begin
    err := SPECK64_Encryption(IV, K64, keySizeWordsN, d);
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Move(d[0], IV[0], 16 * SizeOf(Byte));
    for i := 0 to 15 do
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
      d[i] := AnsiChar(Byte(c[i]) xor Byte(d[i]));
{$ELSE}
      d[i] := c[i] xor d[i];
{$IFEND}
  end;
end;

procedure TSPECKEncryption.EncryptStream(s: TStream; var o: TStream);
var
  err, K, i, t, blockSize, nbReadChar, blockLength, PLength: Integer;
  K16: TSPECKKey16;
  K24: TSPECKKey24;
  K32: TSPECKKey32;
  K48: TSPECKKey48;
  K64: TSPECKKey64;
  nbBlocksRead: Int64;
  P, IVc, c: TSPECKBlock;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  key, IV: PAnsiChar;
  Buffer: array [0 .. 160000] of AnsiChar;
  outputBuffer: array [0 .. 160000] of AnsiChar;
{$ELSE}
  key, IV: PByte;
  Buffer: array [0 .. 160000] of Byte;
  outputBuffer: array [0 .. 160000] of Byte;
{$IFEND}
  conv: TConvert;
begin
  s.Position := 0;
  o.Position := 0;
  SetProgress(0);
  conv := TConvert.Create;
  try
    key := conv.StringToBufferA(FKey, noUni);
    IV := conv.StringToBufferA(FIV, noUni);
  finally
    conv.free;
  end;

  P := GetMemory(16);
  IVc := GetMemory(16);
  c := GetMemory(16);

  err := 0;

  K16 := nil;
  K24 := nil;
  K32 := nil;
  K48 := nil;
  K64 := nil;

  case self.FWordSizeBits of
    wsb16:
      begin
        K16 := GetMemory(22 * SizeOf(Word));
        err := SPECK16_KeyExpansion(key, self.FKeySizeWordsNum, K16);
      end;
    wsb24:
      begin
        K24 := GetMemory(23 * SizeOf(PByte));
        for i := 0 to 22 do
          K24[i] := GetMemory(3);
        err := SPECK24_KeyExpansion(key, self.FKeySizeWordsNum, K24);
      end;
    wsb32:
      begin
        K32 := GetMemory(27 * SizeOf(Integer));
        err := SPECK32_KeyExpansion(key, self.FKeySizeWordsNum, K32);
      end;
    wsb48:
      begin
        K48 := GetMemory(23 * SizeOf(PWord));
        for i := 0 to 28 do
          K48[i] := GetMemory(3);
        err := SPECK48_KeyExpansion(key, self.FKeySizeWordsNum, K48);
      end;
    wsb64:
      begin
        K64 := GetMemory(34 * 2 * SizeOf(Integer));
        err := SPECK64_KeyExpansion(key, self.FKeySizeWordsNum, K64);
      end;
  end;
  if err <> 0 then
    raise ECryptoPack.Create(err);

  try
    PLength := FWordSizeBitsNum div 4;
    blockLength := PLength * 10000 * SizeOf(Buffer[0]);
    FillChar(Buffer, PLength * 10000, 0);
    FillChar(outputBuffer, PLength * 10000, 0);

    if self.paddingMode = TSPECKPaddingMode.nopadding then
    begin
      if s.size mod PLength <> 0 then
        raise ECryptoPack.Create(-211);
    end;

    if self.FType <> TSPECKType.stECB then
    begin
      if self.FIVMode = rand then
      begin
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
        err := RandomBuffer(PLength, IVc);
{$ELSE}
        err := RandomBuffer(PLength, @IVc[0]);
{$IFEND}
        if err <> 0 then
          raise ECryptoPack.Create(err);
        o.Write(IVc^, PLength);
      end
      else
        Move(IV[0], IVc[0], PLength);
    end;

    nbBlocksRead := 0;
    nbReadChar := blockLength;
    while nbReadChar = blockLength do
    begin
      nbReadChar := s.Read(Buffer[0], min(blockLength, s.size));
      if nbReadChar = blockLength then
      begin
        nbBlocksRead := nbBlocksRead + nbReadChar;
        K := 0;
        while K < blockLength do
        begin
          for i := K to K + PLength - 1 do
          begin
            P[i - K] := Buffer[i];
          end;
          case self.FWordSizeBits of
            wsb16:
              SPECKEncryptionLoop16(P, c, K16, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb24:
              SPECKEncryptionLoop24(P, c, K24, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb32:
              SPECKEncryptionLoop32(P, c, K32, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb48:
              SPECKEncryptionLoop48(P, c, K48, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb64:
              SPECKEncryptionLoop64(P, c, K64, self.FKeySizeWordsNum,
                self.FType, IVc);
          end;
          for i := K to K + PLength - 1 do
          begin
            outputBuffer[i] := c[i - K];
          end;
          K := K + PLength;
        end;
        o.Write(outputBuffer[0], blockLength * SizeOf(outputBuffer[0]));
        SetProgress((nbBlocksRead * 100) div s.size);
      end;
    end;

    blockSize := 0;
    if self.paddingMode = TSPECKPaddingMode.PKCS7 then
    begin
      t := (nbReadChar div PLength) * PLength;
      blockSize := nbReadChar mod PLength;
    end
    else
      t := nbReadChar;

    FillChar(outputBuffer, blockLength, 0);
    K := 0;
    while K < t do
    begin
      for i := K to K + PLength - 1 do
      begin
        P[i - K] := Buffer[i];
      end;
      case self.FWordSizeBits of
        wsb16:
          SPECKEncryptionLoop16(P, c, K16, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb24:
          SPECKEncryptionLoop24(P, c, K24, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb32:
          SPECKEncryptionLoop32(P, c, K32, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb48:
          SPECKEncryptionLoop48(P, c, K48, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb64:
          SPECKEncryptionLoop64(P, c, K64, self.FKeySizeWordsNum,
            self.FType, IVc);
      end;
      for i := K to K + PLength - 1 do
      begin
        outputBuffer[i] := c[i - K];
      end;
      K := K + PLength;
    end;
    o.Write(outputBuffer, t * SizeOf(outputBuffer[0]));
    nbBlocksRead := nbBlocksRead + t;
    SetProgress((nbBlocksRead * 100) div s.size);
    if self.paddingMode = TSPECKPaddingMode.PKCS7 then
    begin
      FillChar(P^, 16, 0);
      for i := t to t + blockSize - 1 do
        P[i - t] := Buffer[i];
      err := Padding_speckmode(P, PLength, blockSize);
      if err <> 0 then
        raise ECryptoPack.Create(err);
      case self.FWordSizeBits of
        wsb16:
          SPECKEncryptionLoop16(P, c, K16, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb24:
          SPECKEncryptionLoop24(P, c, K24, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb32:
          SPECKEncryptionLoop32(P, c, K32, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb48:
          SPECKEncryptionLoop48(P, c, K48, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb64:
          SPECKEncryptionLoop64(P, c, K64, self.FKeySizeWordsNum,
            self.FType, IVc);
      end;
      o.Write(c^, PLength * SizeOf(c[0]));
      SetProgress(100);
    end;
  finally
    FreeMem(key);
    FreeMem(IV);
    FreeMem(P);
    FreeMem(IVc);
    FreeMem(c);

    case self.FWordSizeBits of
      wsb16:
        FreeMem(K16);
      wsb24:
        FreeMem(K24);
      wsb32:
        FreeMem(K32);
      wsb48:
        FreeMem(K48);
      wsb64:
        FreeMem(K64);
    end;

  end;
end;

procedure TSPECKEncryption.DecryptStream(s: TStream; var o: TStream);
var
  err, K, i, t, nbReadChar, blockLength, CLength: Integer;
  counter, lenstream, nbBlocksRead: Int64;
  K16: TSPECKKey16;
  K24: TSPECKKey24;
  K32: TSPECKKey32;
  K48: TSPECKKey48;
  K64: TSPECKKey64;
  c, IVc, d: TSPECKBlock;
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  key, IV: PAnsiChar;
  Buffer: array [0 .. 160000] of AnsiChar;
  outputBuffer: array [0 .. 160000] of AnsiChar;
{$ELSE}
  key, IV: PByte;
  Buffer: array [0 .. 160000] of Byte;
  outputBuffer: array [0 .. 160000] of Byte;
{$IFEND}
  conv: TConvert;
begin
  s.Position := 0;
  o.Position := 0;
  SetProgress(0);
  conv := TConvert.Create;
  try
    key := conv.StringToBufferA(FKey, noUni);
    IV := conv.StringToBufferA(FIV, noUni);
  finally
    conv.free;
  end;

  K16 := nil;
  K32 := nil;
  K24 := nil;
  K48 := nil;
  K64 := nil;

  err := 0;
  case self.FWordSizeBits of
    wsb16:
      begin
        K16 := GetMemory(22 * SizeOf(Word));
        err := SPECK16_KeyExpansion(key, self.FKeySizeWordsNum, K16);
      end;
    wsb24:
      begin
        K24 := GetMemory(23 * SizeOf(PByte));
        for i := 0 to 22 do
          K24[i] := GetMemory(3);
        err := SPECK24_KeyExpansion(key, self.FKeySizeWordsNum, K24);
      end;
    wsb32:
      begin
        K32 := GetMemory(27 * SizeOf(Integer));
        err := SPECK32_KeyExpansion(key, self.FKeySizeWordsNum, K32);
      end;
    wsb48:
      begin
        K48 := GetMemory(23 * SizeOf(PWord));
        for i := 0 to 28 do
          K48[i] := GetMemory(3);
        err := SPECK48_KeyExpansion(key, self.FKeySizeWordsNum, K48);
      end;
    wsb64:
      begin
        K64 := GetMemory(34 * SizeOf(Int64));
        err := SPECK64_KeyExpansion(key, self.FKeySizeWordsNum, K64);
      end;
  end;
  if err <> 0 then
    raise ECryptoPack.Create(err);

  c := GetMemory(16);
  IVc := GetMemory(16);
  d := GetMemory(16);

  try
    CLength := FWordSizeBitsNum div 4;
    blockLength := CLength * 10000 * SizeOf(Buffer[0]);
    FillChar(Buffer, CLength * 10000, 0);
    FillChar(outputBuffer, CLength * 10000, 0);

    if self.paddingMode = TSPECKPaddingMode.nopadding then
    begin
      if o.size mod 16 <> 0 then
        raise ECryptoPack.Create(-211);
    end;

    if self.FType <> TSPECKType.stECB then
    begin
      if self.FIVMode = rand then
      begin
        s.Read(IVc^, CLength);
        if self.FPaddingMode = PKCS7 then
          lenstream := s.size - CLength * 2
        else
          lenstream := s.size - CLength - 1;
      end
      else
      begin
        Move(IV[0], IVc[0], CLength);
        if self.FPaddingMode = PKCS7 then
          lenstream := s.size - CLength
        else
          lenstream := s.size - 1;
      end;
    end
    else
    begin
      if self.FPaddingMode = PKCS7 then
        lenstream := s.size - CLength
      else
        lenstream := s.size - 1;
    end;
    nbReadChar := blockLength;
    counter := 0;
    nbBlocksRead := 0;
    while counter <= lenstream do
    begin
      nbReadChar := s.Read(Buffer[0], min(blockLength, s.size));
      counter := counter + nbReadChar;
      if (nbReadChar = blockLength) and (counter <= lenstream) then
      begin
        K := 0;
        nbBlocksRead := nbBlocksRead + nbReadChar;
        while K < blockLength do
        begin
          for i := K to K + CLength - 1 do
          begin
            c[i - K] := Buffer[i];
          end;
          case self.FWordSizeBits of
            wsb16:
              SPECKDecryptionLoop16(c, d, K16, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb24:
              SPECKDecryptionLoop24(c, d, K24, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb32:
              SPECKDecryptionLoop32(c, d, K32, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb48:
              SPECKDecryptionLoop48(c, d, K48, self.FKeySizeWordsNum,
                self.FType, IVc);
            wsb64:
              SPECKDecryptionLoop64(c, d, K64, self.FKeySizeWordsNum,
                self.FType, IVc);
          end;
          for i := K to K + CLength - 1 do
          begin
            outputBuffer[i] := d[i - K];
          end;
          K := K + CLength;
        end;
        o.Write(outputBuffer[0], blockLength * SizeOf(outputBuffer[0]));
        SetProgress((nbBlocksRead * 100) div s.size);
      end;
    end;

    if self.FPaddingMode = PKCS7 then
      t := nbReadChar - CLength
    else
      t := nbReadChar;

    FillChar(outputBuffer, blockLength, 0);
    K := 0;
    while K < t do
    begin
      for i := K to K + CLength - 1 do
      begin
        c[i - K] := Buffer[i];
      end;
      case self.FWordSizeBits of
        wsb16:
          SPECKDecryptionLoop16(c, d, K16, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb24:
          SPECKDecryptionLoop24(c, d, K24, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb32:
          SPECKDecryptionLoop32(c, d, K32, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb48:
          SPECKDecryptionLoop48(c, d, K48, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb64:
          SPECKDecryptionLoop64(c, d, K64, self.FKeySizeWordsNum,
            self.FType, IVc);
      end;
      for i := K to K + CLength - 1 do
      begin
        outputBuffer[i] := d[i - K];
      end;
      K := K + CLength;
    end;
    nbBlocksRead := nbBlocksRead + t;
    SetProgress((nbBlocksRead * 100) div s.size);
    o.Write(outputBuffer, t * SizeOf(outputBuffer[0]));
    if self.paddingMode = TSPECKPaddingMode.PKCS7 then
    begin
      FillChar(c^, 16, 0);
      for i := t to t + CLength - 1 do
        c[i - t] := Buffer[i];

      case self.FWordSizeBits of
        wsb16:
          SPECKDecryptionLoop16(c, d, K16, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb24:
          SPECKDecryptionLoop24(c, d, K24, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb32:
          SPECKDecryptionLoop32(c, d, K32, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb48:
          SPECKDecryptionLoop48(c, d, K48, self.FKeySizeWordsNum,
            self.FType, IVc);
        wsb64:
          SPECKDecryptionLoop64(c, d, K64, self.FKeySizeWordsNum,
            self.FType, IVc);
      end;
      if Byte(d[CLength - 1]) > CLength then
        raise ECryptoPack.Create(-606);
      o.Write(d^, (CLength - Byte(d[CLength - 1])) * SizeOf(d[0]));
      SetProgress(100);
    end;
  finally
    FreeMem(key);
    FreeMem(IV);
    FreeMem(c);
    FreeMem(IVc);
    FreeMem(d);

    case self.FWordSizeBits of
      wsb16:
        FreeMem(K16);
      wsb24:
        FreeMem(K24);
      wsb32:
        FreeMem(K32);
      wsb48:
        FreeMem(K48);
      wsb64:
        FreeMem(K64);
    end;

  end;
end;

procedure TSPECKEncryption.DecryptFileW(s, o: string);
var
  stream1, stream2: TFileStream;
  f: Integer;
begin

  stream1 := TFileStream.Create(s, fmOpenRead);
  try
    if FileExists(o) then
    begin
      System.SysUtils.DeleteFile(o);
      f := FileCreate(o);
      FileClose(f);
    end
    else
    begin
      f := FileCreate(o);
      FileClose(f);
    end;
    stream2 := TFileStream.Create(o, fmOpenReadWrite);
    try
      self.DecryptStream(stream1, TStream(stream2));
    finally
      stream2.free;
    end;
  finally
    stream1.free;
  end;

end;

function TSPECKEncryption.Encrypt(s: string): string;
var
{$IF (defined(MSWINDOWS) or defined(MACOS)) and (not defined(IOS))}
  outputBuffer, inputBuffer, key, IV: PAnsiChar;
{$ELSE}
  outputBuffer, inputBuffer, key, IV: PByte;
{$IFEND}
  outputSizeBytes: PInteger;
  err, msgLen, temp: Integer;
  P: pointer;
  conv: TConvert;
  padd, IVIn: Boolean;
begin
  conv := TConvert.Create(OutputFormat);
  inputBuffer := conv.StringToBuffer(s, FUni, msgLen);
  case FPaddingMode of
    PKCS7:
      begin
        case FIVMode of
          rand:
            begin
              temp := Integer((msgLen DIV (FWordSizeBitsNum DIV 4) + 2) *
                (FWordSizeBitsNum DIV 4));
            end;
          userdefined:
            begin
              temp := Integer((msgLen DIV (FWordSizeBitsNum DIV 4) + 1) *
                (FWordSizeBitsNum DIV 4));
            end;
        end;
      end;
    nopadding:
      begin
        case FIVMode of
          rand:
            begin
              temp := Integer((msgLen DIV (FWordSizeBitsNum DIV 4) + 1) *
                (FWordSizeBitsNum DIV 4));
            end;
          userdefined:
            begin
              temp := Integer((msgLen DIV (FWordSizeBitsNum DIV 4)) *
                (FWordSizeBitsNum DIV 4));
            end;
        end;
      end;
  end;
  P := @temp;
  outputSizeBytes := PInteger(P);
  outputBuffer := GetMemory(outputSizeBytes^);
  key := conv.StringToBufferA(FKey, noUni);
  if FIVMode = userdefined then
    IV := conv.StringToBufferA(FIV, noUni)
  else
    IV := nil;
  try
    padd := (FPaddingMode = TSPECKPaddingMode.PKCS7);
    IVIn := (FIVMode = TSPECKIVMode.rand);
    err := SPECK_EncryptBinaryBuffer(inputBuffer, msgLen, outputBuffer,
      outputSizeBytes, key, FWordSizeBitsNum, FKeySizeWordsNum, Integer(IVIn),
      IV, Integer(padd), Integer(FType));
    if err <> 0 then
      raise ECryptoPack.Create(err);
    Result := conv.CharToFormat(outputBuffer, outputSizeBytes^);

  finally
    FreeMem(inputBuffer);
    FreeMem(outputBuffer);
    FreeMem(key);
    FreeMem(IV);
    conv.free;
  end;

end;

procedure TSPECKEncryption.SetKey(const Value: string);
begin
  if Value <> '' then
  begin
    if (Length(Value) <> ((FWordSizeBitsNum * FKeySizeWordsNum) div 8)) then
      raise ECryptoPack.Create('Keylength incorrect, must be ' +
        inttostr((FWordSizeBitsNum * FKeySizeWordsNum) div 8));
    FKey := Value;
  end;
end;

procedure TSPECKEncryption.SetProgress(P: Integer);
begin
  if P <> FProgress then
  begin
    FProgress := P;
    if Assigned(FOnChange) then
      FOnChange(self);
  end;
end;

procedure TSPECKEncryption.SetIV(const Value: string);
begin
  if Value <> '' then
  begin
    if (Length(Value) <> FWordSizeBitsNum div 4) then
      raise ECryptoPack.Create('IV length incorrect, must be ' +
        inttostr(FWordSizeBitsNum div 4));
    FIV := Value;
  end;
end;

procedure TSPECKEncryption.SetWordSizeBits(const Value: TSPECKWordSizeBits);
begin
  FWordSizeBits := Value;
  case FWordSizeBits of
    wsb16:
      FWordSizeBitsNum := 16;
    wsb24:
      FWordSizeBitsNum := 24;
    wsb32:
      FWordSizeBitsNum := 32;
    wsb48:
      FWordSizeBitsNum := 48;
    wsb64:
      FWordSizeBitsNum := 64;
  end;
end;

procedure TSPECKEncryption.SetKeySizeWords(const Value: TSPECKKeySizeWords);
begin
  FKeySizeWords := Value;
  case FKeySizeWords of
    ksw2:
      begin
        if (FWordSizeBits <> wsb48) and (FWordSizeBits <> wsb64) then
          raise ECryptoPack.Create
            ('KeySizeWords = 2 can be used only with wordSizeBits = 48 or wordSizeBits = 64')
        else
          FKeySizeWordsNum := 2;
      end;
    ksw3:
      begin
        if (FWordSizeBits = wsb16) then
          raise ECryptoPack.Create
            ('KeySizeWords = 3 can be used only with wordSizeBits = 24 or wordSizeBits = 32 or wordSizeBits = 48 or wordSizeBits = 64')
        else
          FKeySizeWordsNum := 3;
      end;
    ksw4:
      begin
        if (FWordSizeBits = wsb48) then
          raise ECryptoPack.Create
            ('KeySizeWords = 4 can be used only with wordSizeBits = 16 or wordSizeBits = 24 or wordSizeBits = 32 or wordSizeBits = 64')
        else
          FKeySizeWordsNum := 4;
      end;
  end;
end;

{$IF (defined(WIN32) or defined(MACOS32)) and (not defined(IOS))}

procedure __llshl;
asm
  jmp System.@_llshl
end;

procedure __llushr;
asm
  jmp System.@_llushr
end;

procedure __llmul;
asm
  jmp System.@_llmul
end;

procedure __llumod;
asm
  jmp System.@_llumod
end;

procedure __lludiv;
asm
  jmp System.@_lludiv
end;

procedure __lldiv;
asm
  jmp System.@_lldiv
end;

procedure __aulldiv; external 'ntdll' name '_aulldiv';

procedure __alldiv; external 'ntdll' name '_alldiv';

procedure __aullrem; external 'ntdll' name '_aullrem';

procedure __chkstk; external 'ntdll' name '_chkstk';

function _wcscpy(destination: PChar; source: PChar): PChar;
  external msvcrt name 'wcscpy';

{$IFEND}
{$IF defined(WIN64)}
procedure __chkstk; external 'ntdll';
{$IFEND}
{$IF defined(WIN32)}
procedure __chkstk_noalloc; external 'ntdll';
{$IFEND}

end.
